from typing import Any

import pandas as pd
import pyarrow as pa
import ray
import ray.data as rd
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.arrow_dataframe import _build_empty_arrow
from fugue.dataframe.utils import (
    get_column_names,
    rename,
)
from fugue_test.dataframe_suite import DataFrameTests
from pytest import raises
from triad import Schema

from fugue_ray import RayDataFrame


class RayDataFrameTests(DataFrameTests.Tests):
    @classmethod
    def setUpClass(cls):
        ray.init(num_cpus=2)

    @classmethod
    def tearDownClass(cls):
        ray.shutdown()

    def df(self, data: Any = None, schema: Any = None) -> RayDataFrame:
        return RayDataFrame(data, schema)

    def test_ray_init(self):
        df = RayDataFrame(schema="a:str,b:int")
        assert df.empty
        assert df.schema == "a:str,b:int"
        assert df.is_bounded

        df = RayDataFrame(pd.DataFrame([], columns=["a", "b"]), schema="a:str,b:int")
        assert df.empty
        assert df.schema == "a:str,b:int"
        assert df.is_bounded

        pdf = pd.DataFrame(dict(a=[1.1, 2.2], b=["a", "b"]))
        pdf = pdf[pdf.a < 0]
        df = RayDataFrame(pdf)
        assert df.empty
        assert df.schema == "a:double,b:str"
        assert df.is_bounded

        df = RayDataFrame([], "x:str,y:double")
        assert df.empty
        assert not df.is_local
        assert df.is_bounded

        df = RayDataFrame(ArrayDataFrame([], "x:str,y:double"))
        assert df.empty
        assert not df.is_local
        assert df.is_bounded

        df = RayDataFrame(_build_empty_arrow(Schema("x:str,y:double")))
        assert df.empty
        assert df.schema == "x:str,y:double"
        assert df.is_bounded

        ddf = RayDataFrame(df)
        assert ddf.empty
        assert ddf.schema == "x:str,y:double"

        ddf = RayDataFrame(df.native, "x:str,y:double")
        assert ddf.empty
        assert ddf.schema == "x:str,y:double"

        ddf = RayDataFrame(df[["x"]].native, "x:str")
        assert ddf.empty
        assert ddf.schema == "x:str"

        raises(Exception, lambda: RayDataFrame(123))

        raises(NotImplementedError, lambda: RayDataFrame(rd.from_items([1, 2])))

    def test_ray_cast(self):
        data = [["a", "1"], ["b", "2"]]
        df = RayDataFrame(data, "a:str,b:str")
        assert [["a", "1"], ["b", "2"]] == df.as_array(type_safe=True)
        data = [["a", 1], ["b", 2]]
        df = RayDataFrame(data, "a:str,b:int")
        assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
        df = RayDataFrame(data, "a:str,b:double")
        assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)

        ddf = RayDataFrame(df.native)
        assert [["a", 1.0], ["b", 2.0]] == ddf.as_array(type_safe=True)

        df = RayDataFrame(df.as_pandas(), "a:str,b:double")
        assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
        df = RayDataFrame(df.as_pandas()["b"])
        assert [[1.0], [2.0]] == df.as_array(type_safe=True)

    def test_ray_num_partitions(self):
        rdf = rd.from_pandas(pd.DataFrame(dict(a=range(10))))
        df = RayDataFrame(rdf.repartition(5))
        assert 5 == df.num_partitions

    def _test_get_column_names(self):
        df = rd.from_pandas(pd.DataFrame([[0, 10, 20]], columns=["0", "1", "2"]))
        assert get_column_names(df) == ["0", "1", "2"]

        df = rd.from_arrow(
            pa.Table.from_pandas(pd.DataFrame([[0, 10, 20]], columns=["0", "1", "2"]))
        )
        assert get_column_names(df) == ["0", "1", "2"]

    def _test_rename(self):
        rdf = rd.from_pandas(pd.DataFrame([[0, 10, 20]], columns=["a", "b", "c"]))
        df = rename(rdf, {})
        assert isinstance(df, rd.Dataset)
        assert get_column_names(df) == ["a", "b", "c"]

        pdf = rd.from_pandas(pd.DataFrame([[0, 10, 20]], columns=["0", "1", "2"]))
        df = rename(pdf, {"0": "_0", "1": "_1", "2": "_2"})
        assert isinstance(df, rd.Dataset)
        assert get_column_names(df) == ["_0", "_1", "_2"]


class NativeRayDataFrameTests(DataFrameTests.NativeTests):
    @classmethod
    def setUpClass(cls):
        ray.init(num_cpus=2)

    @classmethod
    def tearDownClass(cls):
        ray.shutdown()

    def df(self, data: Any = None, schema: Any = None):
        res = RayDataFrame(data, schema)
        # native ray dataset can't handle the schema when empty
        return res if res.empty else res.native

    def to_native_df(self, pdf: pd.DataFrame) -> Any:
        return rd.from_pandas(pdf)



import sys
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from uuid import uuid4

from adagio.specs import WorkflowSpec
from triad import (
    ParamDict,
    Schema,
    SerializableRLock,
    assert_or_throw,
    extensible_class,
)

from fugue._utils.exception import modify_traceback
from fugue.collections.partition import PartitionSpec
from fugue.collections.sql import StructuredRawSQL
from fugue.collections.yielded import Yielded
from fugue.column import ColumnExpr
from fugue.column import SelectColumns as ColumnsSelect
from fugue.column import all_cols, col, lit
from fugue.constants import (
    _FUGUE_GLOBAL_CONF,
    FUGUE_CONF_WORKFLOW_AUTO_PERSIST,
    FUGUE_CONF_WORKFLOW_AUTO_PERSIST_VALUE,
    FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE,
    FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT,
    FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE,
    FUGUE_SQL_DEFAULT_DIALECT,
)
from fugue.dataframe import DataFrame, LocalBoundedDataFrame, YieldedDataFrame
from fugue.dataframe.api import is_df
from fugue.dataframe.dataframes import DataFrames
from fugue.exceptions import FugueWorkflowCompileError, FugueWorkflowError
from fugue.execution.api import engine_context
from fugue.extensions._builtins import (
    Aggregate,
    AlterColumns,
    AssertEqual,
    AssertNotEqual,
    Assign,
    CreateData,
    Distinct,
    DropColumns,
    Dropna,
    Fillna,
    Filter,
    Load,
    Rename,
    RunJoin,
    RunOutputTransformer,
    RunSetOperation,
    RunSQLSelect,
    RunTransformer,
    Sample,
    Save,
    SaveAndUse,
    Select,
    SelectColumns,
    Show,
    Take,
    Zip,
)
from fugue.extensions.transformer.convert import _to_output_transformer, _to_transformer
from fugue.rpc import to_rpc_handler
from fugue.rpc.base import EmptyRPCHandler
from fugue.workflow._checkpoint import StrongCheckpoint, WeakCheckpoint
from fugue.workflow._tasks import Create, FugueTask, Output, Process
from fugue.workflow._workflow_context import FugueWorkflowContext

_DEFAULT_IGNORE_ERRORS: List[Any] = []

TDF = TypeVar("TDF", bound="WorkflowDataFrame")


@extensible_class
class WorkflowDataFrame(DataFrame):
    """It represents the edges in the graph constructed by :class:`~.FugueWorkflow`.
    In Fugue, we use DAG to represent workflows, and the edges are strictly
    dataframes. DAG construction and execution are different steps, this class is
    used in the construction step. Although it inherits from
    :class:`~fugue.dataframe.dataframe.DataFrame`, it's not concerete data. So a
    lot of the operations are not allowed. If you want to obtain the concrete
    Fugue :class:`~fugue.dataframe.dataframe.DataFrame`, use :meth:`~.compute()`
    to execute the workflow.

    Normally, you don't construct it by yourself, you will just use the methods of it.

    :param workflow: the parent workflow it belongs to
    :param task: the task that generates this dataframe
    """

    def __init__(self, workflow: "FugueWorkflow", task: FugueTask):
        super().__init__("_0:int")
        self._workflow = workflow
        self._task = task

    def spec_uuid(self) -> str:
        """UUID of its task spec"""
        return self._task.__uuid__()

    @property
    def native(self) -> Any:  # pragma: no cover
        raise NotImplementedError

    def native_as_df(self) -> Any:  # pragma: no cover
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Name of its task spec"""
        return self._task.name

    @property
    def workflow(self) -> "FugueWorkflow":
        """The parent workflow"""
        return self._workflow

    @property
    def result(self) -> DataFrame:
        """The concrete DataFrame obtained from :meth:`~.compute()`.
        This property will not trigger compute again, but compute should
        have been called earlier and the result is cached.
        """
        return self.workflow.get_result(self)

    @property
    def partition_spec(self) -> PartitionSpec:
        """The partition spec set on the dataframe for next steps to use

        .. admonition:: Examples

            .. code-block:: python

                dag = FugueWorkflow()
                df = dag.df([[0],[1]], "a:int")
                assert df.partition_spec.empty
                df2 = df.partition(by=["a"])
                assert df.partition_spec.empty
                assert df2.partition_spec == PartitionSpec(by=["a"])
        """
        return self.metadata.get("pre_partition", PartitionSpec())

    def compute(self, *args, **kwargs) -> DataFrame:
        """Trigger the parent workflow to
        :meth:`~fugue.workflow.workflow.FugueWorkflow.run` and to generate and cache
        the result dataframe this instance represent.

        .. admonition:: Examples

            >>> df = FugueWorkflow().df([[0]],"a:int").transform(a_transformer)
            >>> df.compute().as_pandas()  # pandas dataframe
            >>> df.compute(SparkExecutionEngine).native  # spark dataframe

        .. note::

            Consider using :meth:`fugue.workflow.workflow.FugueWorkflow.run` instead.
            Because this method actually triggers the entire workflow to run, so it may
            be confusing to use this method because extra time may be taken to compute
            unrelated dataframes.

            .. code-block:: python

                dag = FugueWorkflow()
                df1 = dag.df([[0]],"a:int").transform(a_transformer)
                df2 = dag.df([[0]],"b:int")

                dag.run(SparkExecutionEngine)
                df1.result.show()
                df2.result.show()
        """
        # TODO: it computes entire graph
        self.workflow.run(*args, **kwargs)
        return self.result

    def process(
        self: TDF,
        using: Any,
        schema: Any = None,
        params: Any = None,
        pre_partition: Any = None,
    ) -> TDF:
        """Run a processor on this dataframe. It's a simple wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.process`

        Please read the
        :doc:`Processor Tutorial <tutorial:tutorials/extensions/processor>`

        :param using: processor-like object, if it is a string, then it must be
          the alias of a registered processor
        :param schema: |SchemaLikeObject|, defaults to None. The processor
          will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
        :param params: |ParamsLikeObject| to run the processor, defaults to None.
          The processor will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None.
          The processor will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.partition_spec`
        :return: result dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        if pre_partition is None:
            pre_partition = self.partition_spec
        df = self.workflow.process(
            self, using=using, schema=schema, params=params, pre_partition=pre_partition
        )
        return self._to_self_type(df)

    def output(self, using: Any, params: Any = None, pre_partition: Any = None) -> None:
        """Run a outputter on this dataframe. It's a simple wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.output`

        Please read the
        :doc:`Outputter Tutorial <tutorial:tutorials/extensions/outputter>`

        :param using: outputter-like object, if it is a string, then it must be
          the alias of a registered outputter
        :param params: |ParamsLikeObject| to run the outputter, defaults to None.
          The outputter will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None.
          The outputter will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.partition_spec`
        """
        if pre_partition is None:
            pre_partition = self.partition_spec
        self.workflow.output(
            self, using=using, params=params, pre_partition=pre_partition
        )

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:  # pragma: no cover
        raise NotImplementedError

    def show(
        self,
        n: int = 10,
        with_count: bool = False,
        title: Optional[str] = None,
        best_width: int = 100,
    ) -> None:
        """Show the dataframe.
        See
        :ref:`examples <tutorial:tutorials/advanced/dag:initialize a workflow>`.

        :param n: max number of rows, defaults to 10
        :param with_count: whether to show total count, defaults to False
        :param title: title to display on top of the dataframe, defaults to None
        :param best_width: max width for the output table, defaults to 100

        .. note::

            * When you call this method, it means you want the dataframe to be
              printed when the workflow executes. So the dataframe won't show until
              you run the workflow.
            * When ``with_count`` is True, it can trigger expensive calculation for
              a distributed dataframe. So if you call this function directly, you may
              need to :meth:`~.persist` the dataframe. Or you can turn on
              :ref:`tutorial:tutorials/advanced/useful_config:auto persist`
        """
        # TODO: best_width is not used
        self.workflow.show(self, n=n, with_count=with_count, title=title)

    def assert_eq(self, *dfs: Any, **params: Any) -> None:
        """Wrapper of :meth:`fugue.workflow.workflow.FugueWorkflow.assert_eq` to
        compare this dataframe with other dataframes.

        :param dfs: |DataFramesLikeObject|
        :param digits: precision on float number comparison, defaults to 8
        :param check_order: if to compare the row orders, defaults to False
        :param check_schema: if compare schemas, defaults to True
        :param check_content: if to compare the row values, defaults to True
        :param no_pandas: if true, it will compare the string representations of the
          dataframes, otherwise, it will convert both to pandas dataframe to compare,
          defaults to False

        :raises AssertionError: if not equal
        """
        self.workflow.assert_eq(self, *dfs, **params)

    def assert_not_eq(self, *dfs: Any, **params: Any) -> None:
        """Wrapper of :meth:`fugue.workflow.workflow.FugueWorkflow.assert_not_eq` to
        compare this dataframe with other dataframes.

        :param dfs: |DataFramesLikeObject|
        :param digits: precision on float number comparison, defaults to 8
        :param check_order: if to compare the row orders, defaults to False
        :param check_schema: if compare schemas, defaults to True
        :param check_content: if to compare the row values, defaults to True
        :param no_pandas: if true, it will compare the string representations of the
          dataframes, otherwise, it will convert both to pandas dataframe to compare,
          defaults to False

        :raises AssertionError: if any dataframe is equal to the first dataframe
        """
        self.workflow.assert_not_eq(self, *dfs, **params)

    def select(
        self: TDF,
        *columns: Union[str, ColumnExpr],
        where: Optional[ColumnExpr] = None,
        having: Optional[ColumnExpr] = None,
        distinct: bool = False,
    ) -> TDF:
        """The functional interface for SQL select statement

        :param columns: column expressions, for strings they will represent
          the column names
        :param where: ``WHERE`` condition expression, defaults to None
        :param having: ``having`` condition expression, defaults to None. It
          is used when ``cols`` contains aggregation columns, defaults to None
        :param distinct: whether to return distinct result, defaults to False
        :return: the select result as a new dataframe

        .. admonition:: New Since
            :class: hint

            **0.6.0**

        .. attention::

            This interface is experimental, it's subjected to change in new versions.

        .. seealso::

            Please find more expression examples in :mod:`fugue.column.sql` and
            :mod:`fugue.column.functions`

        .. admonition:: Examples

            .. code-block:: python

                import fugue.column.functions as f
                from fugue import FugueWorkflow

                dag = FugueWorkflow()
                df = dag.df(pandas_df)

                # select existed and new columns
                df.select("a","b",lit(1,"another")))
                df.select("a",(col("b")+lit(1)).alias("x"))

                # select distinct
                df.select("a","b",lit(1,"another")),distinct=True)

                # aggregation
                # SELECT COUNT(DISTINCT *) AS x FROM df
                df.select(f.count_distinct(all_cols()).alias("x"))

                # SELECT a, MAX(b+1) AS x FROM df GROUP BY a
                df.select("a",f.max(col("b")+lit(1)).alias("x"))

                # SELECT a, MAX(b+1) AS x FROM df
                #   WHERE b<2 AND a>1
                #   GROUP BY a
                #   HAVING MAX(b+1)>0
                df.select(
                    "a",f.max(col("b")+lit(1)).alias("x"),
                    where=(col("b")<2) & (col("a")>1),
                    having=f.max(col("b")+lit(1))>0
                )
        """

        def _to_col(s: str) -> ColumnExpr:
            return col(s) if s != "*" else all_cols()

        sc = ColumnsSelect(
            *[_to_col(x) if isinstance(x, str) else x for x in columns],
            arg_distinct=distinct,
        )
        df = self.workflow.process(
            self, using=Select, params=dict(columns=sc, where=where, having=having)
        )
        return self._to_self_type(df)

    def filter(self: TDF, condition: ColumnExpr) -> TDF:
        """Filter rows by the given condition

        :param df: the dataframe to be filtered
        :param condition: (boolean) column expression
        :return: a new filtered dataframe

        .. admonition:: New Since
            :class: hint

            **0.6.0**

        .. seealso::

            Please find more expression examples in :mod:`fugue.column.sql` and
            :mod:`fugue.column.functions`

        .. admonition:: Examples

            .. code-block:: python

                import fugue.column.functions as f
                from fugue import FugueWorkflow

                dag = FugueWorkflow()
                df = dag.df(pandas_df)

                df.filter((col("a")>1) & (col("b")=="x"))
                df.filter(f.coalesce(col("a"),col("b"))>1)
        """
        df = self.workflow.process(self, using=Filter, params=dict(condition=condition))
        return self._to_self_type(df)

    def assign(self: TDF, *args: ColumnExpr, **kwargs: Any) -> TDF:
        """Update existing columns with new values and add new columns

        :param df: the dataframe to set columns
        :param args: column expressions
        :param kwargs: column expressions to be renamed to the argument names,
          if a value is not `ColumnExpr`, it will be treated as a literal
        :return: a new dataframe with the updated values

        .. tip::

            This can be used to cast data types, alter column values or add new
            columns. But you can't use aggregation in columns.

        .. admonition:: New Since
            :class: hint

            **0.6.0**

        .. seealso::

            Please find more expression examples in :mod:`fugue.column.sql` and
            :mod:`fugue.column.functions`

        .. admonition:: Examples

            .. code-block:: python

                from fugue import FugueWorkflow

                dag = FugueWorkflow()
                df = dag.df(pandas_df)

                # add/set 1 as column x
                df.assign(lit(1,"x"))
                df.assign(x=1)

                # add/set x to be a+b
                df.assign((col("a")+col("b")).alias("x"))
                df.assign(x=col("a")+col("b"))

                # cast column a data type to double
                df.assign(col("a").cast(float))

                # cast + new columns
                df.assign(col("a").cast(float),x=1,y=col("a")+col("b"))
        """
        kv: List[ColumnExpr] = [
            v.alias(k) if isinstance(v, ColumnExpr) else lit(v).alias(k)
            for k, v in kwargs.items()
        ]
        df = self.workflow.process(
            self, using=Assign, params=dict(columns=list(args) + kv)
        )
        return self._to_self_type(df)

    def aggregate(self: TDF, *agg_cols: ColumnExpr, **kwagg_cols: ColumnExpr) -> TDF:
        """Aggregate on dataframe

        :param df: the dataframe to aggregate on
        :param agg_cols: aggregation expressions
        :param kwagg_cols: aggregation expressions to be renamed to the argument names
        :return: the aggregated result as a dataframe

        .. admonition:: New Since
            :class: hint

            **0.6.0**

        .. seealso::

            Please find more expression examples in :mod:`fugue.column.sql` and
            :mod:`fugue.column.functions`

        .. admonition:: Examples

            .. code-block:: python

                import fugue.column.functions as f

                # SELECT MAX(b) AS b FROM df
                df.aggregate(f.max(col("b")))

                # SELECT a, MAX(b) AS x FROM df GROUP BY a
                df.partition_by("a").aggregate(f.max(col("b")).alias("x"))
                df.partition_by("a").aggregate(x=f.max(col("b")))
        """
        columns: List[ColumnExpr] = list(agg_cols) + [
            v.alias(k) for k, v in kwagg_cols.items()
        ]
        df = self.workflow.process(
            self,
            using=Aggregate,
            params=dict(columns=columns),
            pre_partition=self.partition_spec,
        )
        return self._to_self_type(df)

    def transform(
        self: TDF,
        using: Any,
        schema: Any = None,
        params: Any = None,
        pre_partition: Any = None,
        ignore_errors: List[Any] = _DEFAULT_IGNORE_ERRORS,
        callback: Any = None,
    ) -> TDF:
        """Transform this dataframe using transformer. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.transform`

        Please read |TransformerTutorial|

        :param using: transformer-like object, if it is a string, then it must be
          the alias of a registered transformer/cotransformer
        :param schema: |SchemaLikeObject|, defaults to None. The transformer
          will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
        :param params: |ParamsLikeObject| to run the processor, defaults to None.
          The transformer will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None. It's
          recommended to use the equivalent wayt, which is to call
          :meth:`~.partition` and then call :meth:`~.transform` without this parameter
        :param ignore_errors: list of exception types the transformer can ignore,
          defaults to empty list
        :param callback: |RPCHandlerLikeObject|, defaults to None
        :return: the transformed dataframe
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            :meth:`~.transform` can be lazy and will return the transformed dataframe,
            :meth:`~.out_transform` is guaranteed to execute immediately (eager) and
            return nothing
        """
        if pre_partition is None:
            pre_partition = self.partition_spec
        df = self.workflow.transform(
            self,
            using=using,
            schema=schema,
            params=params,
            pre_partition=pre_partition,
            ignore_errors=ignore_errors,
            callback=callback,
        )
        return self._to_self_type(df)

    def out_transform(
        self: TDF,
        using: Any,
        params: Any = None,
        pre_partition: Any = None,
        ignore_errors: List[Any] = _DEFAULT_IGNORE_ERRORS,
        callback: Any = None,
    ) -> None:
        """Transform this dataframe using transformer. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.out_transform`

        Please read |TransformerTutorial|

        :param using: transformer-like object, if it is a string, then it must be
          the alias of a registered output transformer/cotransformer
        :param params: |ParamsLikeObject| to run the processor, defaults to None.
          The transformer will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None. It's
          recommended to use the equivalent wayt, which is to call
          :meth:`~.partition` and then call :meth:`~.transform` without this parameter
        :param ignore_errors: list of exception types the transformer can ignore,
          defaults to empty list
        :param callback: |RPCHandlerLikeObject|, defaults to None

        .. note::

            :meth:`~.transform` can be lazy and will return the transformed dataframe,
            :meth:`~.out_transform` is guaranteed to execute immediately (eager) and
            return nothing
        """
        if pre_partition is None:
            pre_partition = self.partition_spec
        self.workflow.out_transform(
            self,
            using=using,
            params=params,
            pre_partition=pre_partition,
            ignore_errors=ignore_errors,
            callback=callback,
        )

    def join(self: TDF, *dfs: Any, how: str, on: Optional[Iterable[str]] = None) -> TDF:
        """Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param how: can accept ``semi``, ``left_semi``, ``anti``, ``left_anti``,
          ``inner``, ``left_outer``, ``right_outer``, ``full_outer``, ``cross``
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        df = self.workflow.join(self, *dfs, how=how, on=on)
        return self._to_self_type(df)

    def inner_join(self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None) -> TDF:
        """INNER Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="inner", on=on)

    def semi_join(self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None) -> TDF:
        """LEFT SEMI Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="semi", on=on)

    def left_semi_join(self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None) -> TDF:
        """LEFT SEMI Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="left_semi", on=on)

    def anti_join(self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None) -> TDF:
        """LEFT ANTI Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="anti", on=on)

    def left_anti_join(self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None) -> TDF:
        """LEFT ANTI Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="left_anti", on=on)

    def left_outer_join(
        self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None
    ) -> TDF:
        """LEFT OUTER Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="left_outer", on=on)

    def right_outer_join(
        self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None
    ) -> TDF:
        """RIGHT OUTER Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="right_outer", on=on)

    def full_outer_join(
        self: TDF, *dfs: Any, on: Optional[Iterable[str]] = None
    ) -> TDF:
        """CROSS Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="full_outer", on=on)

    def cross_join(self: TDF, *dfs: Any) -> TDF:
        """CROSS Join this dataframe with dataframes. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.join`. |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :return: joined dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        return self.join(*dfs, how="cross")

    def union(self: TDF, *dfs: Any, distinct: bool = True) -> TDF:
        """Union this dataframe with ``dfs``.

        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after union,
          default to True
        :return: unioned dataframe

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        df = self.workflow.union(self, *dfs, distinct=distinct)
        return self._to_self_type(df)

    def subtract(self: TDF, *dfs: Any, distinct: bool = True) -> TDF:
        """Subtract ``dfs`` from this dataframe.

        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after subtraction,
          default to True
        :return: subtracted dataframe

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        df = self.workflow.subtract(self, *dfs, distinct=distinct)
        return self._to_self_type(df)

    def intersect(self: TDF, *dfs: Any, distinct: bool = True) -> TDF:
        """Intersect this dataframe with ``dfs``.

        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after intersection,
          default to True
        :return: intersected dataframe

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        df = self.workflow.intersect(self, *dfs, distinct=distinct)
        return self._to_self_type(df)

    def distinct(self: TDF) -> TDF:
        """Get distinct dataframe. Equivalent to ``SELECT DISTINCT * FROM df``

        :return: dataframe with unique records
        """
        df = self.workflow.process(self, using=Distinct)
        return self._to_self_type(df)

    def dropna(
        self: TDF, how: str = "any", thresh: int = None, subset: List[str] = None
    ) -> TDF:
        """Drops records containing NA records

        :param how: 'any' or 'all'. 'any' drops rows that contain any nulls.
          'all' drops rows that contain all nulls.
        :param thresh: int, drops rows that have less than thresh non-null values
        :param subset: list of columns to operate on
        :return: dataframe with incomplete records dropped
        """
        params = dict(how=how, thresh=thresh, subset=subset)
        params = {k: v for k, v in params.items() if v is not None}
        df = self.workflow.process(self, using=Dropna, params=params)
        return self._to_self_type(df)

    def fillna(self: TDF, value: Any, subset: List[str] = None) -> TDF:
        """Fills NA values with replacement values

        :param value: if scalar, fills all columns with same value.
            if dictionary, fills NA using the keys as column names and the
            values as the replacement values.
        :param subset: list of columns to operate on. ignored if value is
            a dictionary

        :return: dataframe with NA records filled
        """
        params = dict(value=value, subset=subset)
        params = {k: v for k, v in params.items() if v is not None}
        df = self.workflow.process(self, using=Fillna, params=params)
        return self._to_self_type(df)

    def sample(
        self: TDF,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> TDF:
        """
        Sample dataframe by number of rows or by fraction

        :param n: number of rows to sample, one and only one of ``n`` and ``fact``
          must be set
        :param frac: fraction [0,1] to sample, one and only one of ``n`` and ``fact``
          must be set
        :param replace: whether replacement is allowed. With replacement,
          there may be duplicated rows in the result, defaults to False
        :param seed: seed for randomness, defaults to None

        :return: sampled dataframe
        """
        params: Dict[str, Any] = dict(replace=replace)
        if seed is not None:
            params["seed"] = seed
        if n is not None:
            params["n"] = n
        if frac is not None:
            params["frac"] = frac
        df = self.workflow.process(self, using=Sample, params=params)
        return self._to_self_type(df)

    def take(self: TDF, n: int, presort: str = None, na_position: str = "last") -> TDF:
        """
        Get the first n rows of a DataFrame per partition. If a presort is defined,
        use the presort before applying take. presort overrides partition_spec.presort

        :param n: number of rows to return
        :param presort: presort expression similar to partition presort
        :param na_position: position of null values during the presort.
            can accept ``first`` or ``last``

        :return: n rows of DataFrame per partition
        """
        params: Dict[str, Any] = dict()
        params["n"] = n
        # Note float is converted to int with triad _get_or
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )
        assert_or_throw(
            na_position in ("first", "last"),
            ValueError("na_position must be either 'first' or 'last'"),
        )
        params["na_position"] = na_position
        if presort is not None:
            params["presort"] = presort

        df = self.workflow.process(
            self, using=Take, pre_partition=self.partition_spec, params=params
        )
        return self._to_self_type(df)

    def weak_checkpoint(self: TDF, lazy: bool = False, **kwargs: Any) -> TDF:
        """Cache the dataframe in memory

        :param lazy: whether it is a lazy checkpoint, defaults to False (eager)
        :param kwargs: paramteters for the underlying execution engine function
        :return: the cached dataframe

        .. note::

            Weak checkpoint in most cases is the best choice for caching a dataframe to
            avoid duplicated computation. However it does not guarantee to break up the
            the compute dependency for this dataframe, so when you have very complicated
            compute, you may encounter issues such as stack overflow. Also, weak
            checkpoint normally caches the dataframe in memory, if memory is a concern,
            then you should consider :meth:`~.strong_checkpoint`
        """
        self._task.set_checkpoint(WeakCheckpoint(lazy=lazy, **kwargs))
        return self

    def strong_checkpoint(
        self: TDF,
        storage_type: str = "file",
        lazy: bool = False,
        partition: Any = None,
        single: bool = False,
        **kwargs: Any,
    ) -> TDF:
        """Cache the dataframe as a temporary file

        :param storage_type: can be either ``file`` or ``table``, defaults to ``file``
        :param lazy: whether it is a lazy checkpoint, defaults to False (eager)
        :param partition: |PartitionLikeObject|, defaults to None.
        :param single: force the output as a single file, defaults to False
        :param kwargs: paramteters for the underlying execution engine function
        :return: the cached dataframe

        .. note::

            Strong checkpoint guarantees the output dataframe compute dependency is
            from the temporary file. Use strong checkpoint only when
            :meth:`~.weak_checkpoint` can't be used.

            Strong checkpoint file will be removed after the execution of the workflow.
        """
        self._task.set_checkpoint(
            StrongCheckpoint(
                storage_type=storage_type,
                obj_id=str(uuid4()),
                deterministic=False,
                permanent=False,
                lazy=lazy,
                partition=partition,
                single=single,
                **kwargs,
            )
        )
        return self

    def deterministic_checkpoint(
        self: TDF,
        storage_type: str = "file",
        lazy: bool = False,
        partition: Any = None,
        single: bool = False,
        namespace: Any = None,
        **kwargs: Any,
    ) -> TDF:
        """Cache the dataframe as a temporary file

        :param storage_type: can be either ``file`` or ``table``, defaults to ``file``
        :param lazy: whether it is a lazy checkpoint, defaults to False (eager)
        :param partition: |PartitionLikeObject|, defaults to None.
        :param single: force the output as a single file, defaults to False
        :param kwargs: paramteters for the underlying execution engine function
        :param namespace: a value to control determinism, defaults to None.
        :return: the cached dataframe

        .. note::

            The difference vs :meth:`~.strong_checkpoint` is that this checkpoint is not
            removed after execution, so it can take effect cross execution if the
            dependent compute logic is not changed.
        """
        self._task.set_checkpoint(
            StrongCheckpoint(
                storage_type=storage_type,
                obj_id=self._task.__uuid__(),
                deterministic=True,
                permanent=True,
                lazy=lazy,
                partition=partition,
                single=single,
                namespace=namespace,
                **kwargs,
            )
        )
        return self

    def yield_file_as(self: TDF, name: str) -> None:
        """Cache the dataframe in file

        :param name: the name of the yielded dataframe

        .. note::

            In only the following cases you can yield file/table:

            * you have not checkpointed (persisted) the dataframe, for example
              ``df.yield_file_as("a")``
            * you have used :meth:`~.deterministic_checkpoint`, for example
              ``df.deterministic_checkpoint().yield_file_as("a")``
            * yield is workflow, compile level logic

            For the first case, the yield will also be a strong checkpoint so
            whenever you yield a dataframe as a file, the dataframe has been saved as a
            file and loaded back as a new dataframe.
        """
        if not self._task.has_checkpoint:
            # the following == a non determinitic, but permanent checkpoint
            self.deterministic_checkpoint(storage_type="file", namespace=str(uuid4()))
        self.workflow._yields[name] = self._task.yielded

    def yield_table_as(self: TDF, name: str) -> None:
        """Cache the dataframe as a table

        :param name: the name of the yielded dataframe

        .. note::

            In only the following cases you can yield file/table:

            * you have not checkpointed (persisted) the dataframe, for example
              ``df.yield_file_as("a")``
            * you have used :meth:`~.deterministic_checkpoint`, for example
              ``df.deterministic_checkpoint().yield_file_as("a")``
            * yield is workflow, compile level logic

            For the first case, the yield will also be a strong checkpoint so
            whenever you yield a dataframe as a file, the dataframe has been saved as a
            file and loaded back as a new dataframe.
        """
        if not self._task.has_checkpoint:
            # the following == a non determinitic, but permanent checkpoint
            self.deterministic_checkpoint(storage_type="table", namespace=str(uuid4()))
        self.workflow._yields[name] = self._task.yielded

    def yield_dataframe_as(self: TDF, name: str, as_local: bool = False) -> None:
        """Yield a dataframe that can be accessed without
        the current execution engine

        :param name: the name of the yielded dataframe
        :param as_local: yield the local version of the dataframe

        .. note::

            When ``as_local`` is True, it can trigger an additional compute
            to do the conversion. To avoid recompute, you should add
            ``persist`` before yielding.
        """
        yielded = YieldedDataFrame(self._task.__uuid__())
        self.workflow._yields[name] = yielded
        self._task.set_yield_dataframe_handler(
            lambda df: yielded.set_value(df), as_local=as_local
        )

    def persist(self: TDF) -> TDF:
        """Persist the current dataframe

        :return: the persisted dataframe
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            ``persist`` can only guarantee the persisted dataframe will be computed
            for only once. However this doesn't mean the backend really breaks up the
            execution dependency at the persisting point. Commonly, it doesn't cause
            any issue, but if your execution graph is long, it may cause expected
            problems for example, stack overflow.

            ``persist`` method is considered as weak checkpoint. Sometimes, it may be
            necessary to use strong checkpint, which is :meth:`~.checkpoint`
        """
        return self.weak_checkpoint(lazy=False)

    def checkpoint(self: TDF, storage_type: str = "file") -> TDF:
        return self.strong_checkpoint(storage_type=storage_type, lazy=False)

    def broadcast(self: TDF) -> TDF:
        """Broadcast the current dataframe

        :return: the broadcasted dataframe
        :rtype: :class:`~.WorkflowDataFrame`
        """
        self._task.broadcast()
        return self

    def partition(self: TDF, *args: Any, **kwargs: Any) -> TDF:
        """Partition the current dataframe. Please read |PartitionTutorial|

        :param args: |PartitionLikeObject|
        :param kwargs: |PartitionLikeObject|
        :return: dataframe with the partition hint
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            Normally this step is fast because it's to add a partition hint
            for the next step.
        """
        res = WorkflowDataFrame(self.workflow, self._task)
        res.reset_metadata({"pre_partition": PartitionSpec(*args, **kwargs)})
        return self._to_self_type(res)

    def partition_by(self: TDF, *keys: str, **kwargs: Any) -> TDF:
        """Partition the current dataframe by keys. Please read |PartitionTutorial|.
        This is a wrapper of :meth:`~.partition`

        :param keys: partition keys
        :param kwargs: |PartitionLikeObject| excluding ``by`` and ``partition_by``
        :return: dataframe with the partition hint
        :rtype: :class:`~.WorkflowDataFrame`
        """
        assert_or_throw(len(keys) > 0, FugueWorkflowCompileError("keys can't be empty"))
        assert_or_throw(
            "by" not in kwargs and "partition_by" not in kwargs,
            FugueWorkflowCompileError("by and partition_by can't be in kwargs"),
        )
        return self.partition(by=keys, **kwargs)

    def per_partition_by(self: TDF, *keys: str) -> TDF:
        """Partition the current dataframe by keys so each physical partition contains
        only one logical partition. Please read |PartitionTutorial|.
        This is a wrapper of :meth:`~.partition`

        :param keys: partition keys
        :return: dataframe that is both logically and physically partitioned by ``keys``
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            This is a hint but not enforced, certain execution engines will not
            respect this hint.
        """
        return self.partition_by(*keys, algo="even")

    def per_row(self: TDF) -> TDF:
        """Partition the current dataframe to one row per partition.
        Please read |PartitionTutorial|. This is a wrapper of :meth:`~.partition`

        :return: dataframe that is evenly partitioned by row count
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            This is a hint but not enforced, certain execution engines will not
            respect this hint.
        """
        return self.partition("per_row")

    def _to_self_type(self: TDF, df: "WorkflowDataFrame") -> TDF:
        return df  # type: ignore

    def drop(  # type: ignore
        self: TDF, columns: List[str], if_exists: bool = False
    ) -> TDF:
        """Drop columns from the dataframe.

        :param columns: columns to drop
        :param if_exists: if setting to True, it will ignore non-existent columns,
          defaults to False
        :return: the dataframe after dropping columns
        :rtype: :class:`~.WorkflowDataFrame`
        """
        df = self.workflow.process(
            self, using=DropColumns, params=dict(columns=columns, if_exists=if_exists)
        )
        return self._to_self_type(df)

    def rename(self: TDF, *args: Any, **kwargs: Any) -> TDF:
        """Rename the dataframe using a mapping dict

        :param args: list of dicts containing rename maps
        :param kwargs: rename map
        :return: a new dataframe with the new names
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            This interface is more flexible than
            :meth:`fugue.dataframe.dataframe.DataFrame.rename`

        .. admonition:: Examples

            >>> df.rename({"a": "b"}, c="d", e="f")
        """
        m: Dict[str, str] = {}
        for a in args:
            m.update(a)
        m.update(kwargs)
        df = self.workflow.process(self, using=Rename, params=dict(columns=m))
        return self._to_self_type(df)

    def alter_columns(self: TDF, columns: Any) -> TDF:
        """Change column types

        :param columns: |SchemaLikeObject|
        :return: a new dataframe with the new column types
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            The output dataframe will not change the order of original schema.

        .. admonition:: Examples

            >>> df.alter_columns("a:int,b;str")
        """
        df = self.workflow.process(
            self, using=AlterColumns, params=dict(columns=columns)
        )
        return self._to_self_type(df)

    def zip(
        self: TDF,
        *dfs: Any,
        how: str = "inner",
        partition: Any = None,
        temp_path: Optional[str] = None,
        to_file_threshold: Any = -1,
    ) -> TDF:
        """Zip this data frame with multiple dataframes together
        with given partition specifications. It's a wrapper of
        :meth:`fugue.workflow.workflow.FugueWorkflow.zip`.

        :param dfs: |DataFramesLikeObject|
        :param how: can accept ``inner``, ``left_outer``, ``right_outer``,
          ``full_outer``, ``cross``, defaults to ``inner``
        :param partition: |PartitionLikeObject|, defaults to None.
        :param temp_path: file path to store the data (used only if the serialized data
          is larger than ``to_file_threshold``), defaults to None
        :param to_file_threshold: file byte size threshold, defaults to -1

        :return: a zipped dataframe
        :rtype: :class:`~.WorkflowDataFrame`

        .. note::

            * ``dfs`` must be list like, the zipped dataframe will be list like
            * ``dfs`` is fine to be empty
            * If you want dict-like zip, use
              :meth:`fugue.workflow.workflow.FugueWorkflow.zip`

        .. seealso::

            Read |CoTransformer| and |ZipComap| for details
        """
        if partition is None:
            partition = self.partition_spec
        df = self.workflow.zip(
            self,
            *dfs,
            how=how,
            partition=partition,
            temp_path=temp_path,
            to_file_threshold=to_file_threshold,
        )
        return self._to_self_type(df)

    def __getitem__(self: TDF, columns: List[Any]) -> TDF:
        df = self.workflow.process(
            self, using=SelectColumns, params=dict(columns=columns)
        )
        return self._to_self_type(df)

    def save(
        self,
        path: str,
        fmt: str = "",
        mode: str = "overwrite",
        partition: Any = None,
        single: bool = False,
        **kwargs: Any,
    ) -> None:
        """Save this dataframe to a persistent storage

        :param path: output path
        :param fmt: format hint can accept ``parquet``, ``csv``, ``json``,
          defaults to None, meaning to infer
        :param mode: can accept ``overwrite``, ``append``, ``error``,
          defaults to "overwrite"
        :param partition: |PartitionLikeObject|, how to partition the
          dataframe before saving, defaults to empty
        :param single: force the output as a single file, defaults to False
        :param kwargs: parameters to pass to the underlying framework

        For more details and examples, read
        :ref:`Save & Load <tutorial:tutorials/advanced/dag:save & load>`.
        """
        if partition is None:
            partition = self.partition_spec
        self.workflow.output(
            self,
            using=Save,
            pre_partition=partition,
            params=dict(path=path, fmt=fmt, mode=mode, single=single, params=kwargs),
        )

    def save_and_use(
        self: TDF,
        path: str,
        fmt: str = "",
        mode: str = "overwrite",
        partition: Any = None,
        single: bool = False,
        **kwargs: Any,
    ) -> TDF:
        """Save this dataframe to a persistent storage and load back to use
        in the following steps

        :param path: output path
        :param fmt: format hint can accept ``parquet``, ``csv``, ``json``,
          defaults to None, meaning to infer
        :param mode: can accept ``overwrite``, ``append``, ``error``,
          defaults to "overwrite"
        :param partition: |PartitionLikeObject|, how to partition the
          dataframe before saving, defaults to empty
        :param single: force the output as a single file, defaults to False
        :param kwargs: parameters to pass to the underlying framework

        For more details and examples, read
        :ref:`Save & Load <tutorial:tutorials/advanced/dag:save & load>`.
        """
        if partition is None:
            partition = self.partition_spec
        df = self.workflow.process(
            self,
            using=SaveAndUse,
            pre_partition=partition,
            params=dict(path=path, fmt=fmt, mode=mode, single=single, params=kwargs),
        )
        return self._to_self_type(df)

    @property
    def schema(self) -> Schema:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    @property
    def is_local(self) -> bool:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def as_local(self) -> DataFrame:  # type: ignore  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def as_local_bounded(self) -> DataFrame:  # type: ignore  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    @property
    def is_bounded(self) -> bool:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    @property
    def empty(self) -> bool:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    @property
    def num_partitions(self) -> int:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def peek_array(self) -> List[Any]:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def count(self) -> int:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:  # pragma: no cover
        """
        :raises NotImplementedError: don't call this method
        """
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def _drop_cols(self: TDF, cols: List[str]) -> DataFrame:  # pragma: no cover
        raise NotImplementedError("WorkflowDataFrame does not support this method")

    def _select_cols(self, keys: List[Any]) -> DataFrame:  # pragma: no cover
        raise NotImplementedError("WorkflowDataFrame does not support this method")


class WorkflowDataFrames(DataFrames):
    """Ordered dictionary of WorkflowDataFrames. There are two modes: with keys
    and without keys. If without key ``_<n>`` will be used as the key
    for each dataframe, and it will be treated as an array in Fugue framework.

    It's immutable, once initialized, you can't add or remove element from it.

    It's a subclass of
    :class:`~fugue.dataframe.dataframes.DataFrames`, but different from
    DataFrames, in the initialization you should always use
    :class:`~fugue.workflow.workflow.WorkflowDataFrame`, and they should all
    come from the same :class:`~fugue.workflow.workflow.FugueWorkflow`.

    .. admonition:: Examples

        .. code-block:: python

            dag = FugueWorkflow()
            df1 = dag.df([[0]],"a:int").transform(a_transformer)
            df2 = dag.df([[0]],"b:int")
            dfs1 = WorkflowDataFrames(df1, df2)  # as array
            dfs2 = WorkflowDataFrames([df1, df2])  # as array
            dfs3 = WorkflowDataFrames(a=df1, b=df2)  # as dict
            dfs4 = WorkflowDataFrames(dict(a=df1, b=df2))  # as dict
            dfs5 = WorkflowDataFrames(dfs4, c=df2)  # copy and update
            dfs5["b"].show()  # how you get element when it's a dict
            dfs1[0].show()  # how you get element when it's an array
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self._parent: Optional["FugueWorkflow"] = None
        super().__init__(*args, **kwargs)

    @property
    def workflow(self) -> "FugueWorkflow":
        """The parent workflow"""
        assert_or_throw(
            self._parent is not None, ValueError("parent workflow is unknown")
        )
        return self._parent  # type: ignore

    def __setitem__(  # type: ignore
        self, key: str, value: WorkflowDataFrame, *args: Any, **kwds: Any
    ) -> None:
        assert_or_throw(
            isinstance(value, WorkflowDataFrame),
            lambda: ValueError(f"{key}:{value} is not WorkflowDataFrame)"),
        )
        if self._parent is None:
            self._parent = value.workflow
        else:
            assert_or_throw(
                self._parent is value.workflow,
                ValueError("different parent workflow detected in dataframes"),
            )
        super().__setitem__(key, value, *args, **kwds)

    def __getitem__(  # pylint: disable=W0235
        self, key: Union[str, int]  # type: ignore
    ) -> WorkflowDataFrame:
        return super().__getitem__(key)  # type: ignore

    def __getattr__(self, name: str) -> Any:  # pragma: no cover
        """The dummy method to avoid PyLint complaint"""
        raise AttributeError(name)


class FugueWorkflowResult(DataFrames):
    """The result object of :meth:`~.FugueWorkflow.run`. Users should not
    construct this object.

    :param DataFrames: yields of the workflow
    """

    def __init__(self, yields: Dict[str, Yielded]):
        self._yields = yields
        super().__init__(
            {k: v.result for k, v in yields.items() if isinstance(v, YieldedDataFrame)}
        )

    @property
    def yields(self) -> Dict[str, Any]:
        return self._yields


@extensible_class
class FugueWorkflow:
    """Fugue Workflow, also known as the Fugue Programming Interface.

    In Fugue, we use DAG to represent workflows, DAG construction and execution
    are different steps, this class is mainly used in the construction step, so all
    things you added to the workflow is **description** and they are not executed
    until you call :meth:`~.run`

    Read
    :ref:`this <tutorial:tutorials/advanced/dag:initialize a workflow>`
    to learn how to initialize it in different ways and pros and cons.
    """

    def __init__(self, compile_conf: Any = None):
        assert_or_throw(
            compile_conf is None or isinstance(compile_conf, (dict, ParamDict)),
            ValueError(
                f"FugueWorkflow no longer takes {type(compile_conf)} as the input"
            ),
        )

        self._lock = SerializableRLock()
        self._spec = WorkflowSpec()
        self._computed = False
        self._graph = _Graph()
        self._yields: Dict[str, Yielded] = {}
        self._compile_conf = ParamDict(
            {**_FUGUE_GLOBAL_CONF, **ParamDict(compile_conf)}
        )
        self._last_df: Optional[WorkflowDataFrame] = None

    @property
    def conf(self) -> ParamDict:
        """Compile time configs"""
        return self._compile_conf

    def spec_uuid(self) -> str:
        """UUID of the workflow spec (`description`)"""
        return self._spec.__uuid__()

    def run(
        self, engine: Any = None, conf: Any = None, **kwargs: Any
    ) -> FugueWorkflowResult:
        """Execute the workflow and compute all dataframes.

        .. note::

            For inputs, please read
            :func:`~.fugue.api.engine_context`

        :param engine: object that can be recognized as an engine, defaults to None
        :param conf: engine config, defaults to None
        :param kwargs: additional parameters to initialize the execution engine
        :return: the result set

        .. admonition:: Examples

            .. code-block:: python

                dag = FugueWorkflow()
                df1 = dag.df([[0]],"a:int").transform(a_transformer)
                df2 = dag.df([[0]],"b:int")

                dag.run(SparkExecutionEngine)
                df1.result.show()
                df2.result.show()

                dag = FugueWorkflow()
                df1 = dag.df([[0]],"a:int").transform(a_transformer)
                df1.yield_dataframe_as("x")

                result = dag.run(SparkExecutionEngine)
                result["x"]  # SparkDataFrame

        Read
        :ref:`this <tutorial:tutorials/advanced/dag:initialize a workflow>`
        to learn how to run in different ways and pros and cons.
        """
        with self._lock:
            with engine_context(engine, engine_conf=conf) as e:
                self._computed = False
                self._workflow_ctx = FugueWorkflowContext(
                    engine=e, compile_conf=self.conf
                )
                try:
                    self._workflow_ctx.run(self._spec, {})
                except Exception as ex:
                    if not self.conf.get_or_throw(
                        FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE, bool
                    ) or sys.version_info < (3, 7):
                        raise
                    conf = self.conf.get_or_throw(
                        FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE, str
                    )
                    pre = [p for p in conf.split(",") if p != ""]
                    if len(pre) == 0:
                        raise

                    # prune by prefix
                    ctb = modify_traceback(
                        sys.exc_info()[2],
                        lambda x: any(x.lower().startswith(xx) for xx in pre),
                    )
                    if ctb is None:  # pragma: no cover
                        raise
                    raise ex.with_traceback(ctb)
                self._computed = True
        return FugueWorkflowResult(self.yields)

    @property
    def yields(self) -> Dict[str, Yielded]:
        return self._yields

    @property
    def last_df(self) -> Optional[WorkflowDataFrame]:
        return self._last_df

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return

    def get_result(self, df: WorkflowDataFrame) -> DataFrame:
        """After :meth:`~.run`, get the result of a dataframe defined in the dag

        :return: a calculated dataframe

        .. admonition:: Examples

            .. code-block:: python

                dag = FugueWorkflow()
                df1 = dag.df([[0]],"a:int")
                dag.run()
                dag.get_result(df1).show()
        """
        assert_or_throw(self._computed, FugueWorkflowError("not computed"))
        return self._workflow_ctx.get_result(id(df._task))

    def create(
        self, using: Any, schema: Any = None, params: Any = None
    ) -> WorkflowDataFrame:
        """Run a creator to create a dataframe.

        Please read the
        :doc:`Creator Tutorial <tutorial:tutorials/extensions/creator>`

        :param using: creator-like object, if it is a string, then it must be
          the alias of a registered creator
        :param schema: |SchemaLikeObject|, defaults to None. The creator
          will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
        :param params: |ParamsLikeObject| to run the creator,
          defaults to None. The creator will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None.
          The creator will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.partition_spec`
        :return: result dataframe
        """
        task = Create(
            creator=CreateData(using)
            if is_df(using) or isinstance(using, Yielded)
            else using,
            schema=schema,
            params=params,
        )
        res = self.add(task)
        self._last_df = res
        return res

    def process(
        self,
        *dfs: Any,
        using: Any,
        schema: Any = None,
        params: Any = None,
        pre_partition: Any = None,
    ) -> WorkflowDataFrame:
        """Run a processor on the dataframes.

        Please read the
        :doc:`Processor Tutorial <tutorial:tutorials/extensions/processor>`

        :param dfs: |DataFramesLikeObject|
        :param using: processor-like object, if it is a string, then it must be
          the alias of a registered processor
        :param schema: |SchemaLikeObject|, defaults to None. The processor
          will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
        :param params: |ParamsLikeObject| to run the processor, defaults to None.
          The processor will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None.
          The processor will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.partition_spec`

        :return: result dataframe
        """
        _dfs = self._to_dfs(*dfs)
        task = Process(
            len(_dfs),
            processor=using,
            schema=schema,
            params=params,
            pre_partition=pre_partition,
            input_names=None if not _dfs.has_key else list(_dfs.keys()),
        )
        if _dfs.has_key:
            res = self.add(task, **_dfs)
        else:
            res = self.add(task, *_dfs.values())
        self._last_df = res
        return res

    def output(
        self, *dfs: Any, using: Any, params: Any = None, pre_partition: Any = None
    ) -> None:
        """Run a outputter on dataframes.

        Please read the
        :doc:`Outputter Tutorial <tutorial:tutorials/extensions/outputter>`

        :param using: outputter-like object, if it is a string, then it must be
          the alias of a registered outputter
        :param params: |ParamsLikeObject| to run the outputter, defaults to None.
          The outputter will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None.
          The outputter will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.partition_spec`
        """
        _dfs = self._to_dfs(*dfs)
        task = Output(
            len(_dfs),
            outputter=using,
            params=params,
            pre_partition=pre_partition,
            input_names=None if not _dfs.has_key else list(_dfs.keys()),
        )
        if _dfs.has_key:
            self.add(task, **_dfs)
        else:
            self.add(task, *_dfs.values())

    def create_data(
        self,
        data: Any,
        schema: Any = None,
        data_determiner: Optional[Callable[[Any], Any]] = None,
    ) -> WorkflowDataFrame:
        """Create dataframe.

        :param data: |DataFrameLikeObject| or :class:`~fugue.workflow.yielded.Yielded`
        :param schema: |SchemaLikeObject|, defaults to None
        :param data_determiner: a function to compute unique id from ``data``
        :return: a dataframe of the current workflow

        .. note::

            By default, the input ``data`` does not affect the determinism of the
            workflow (but ``schema`` and ``etadata`` do), because the amount of compute
            can be unpredictable. But if you want ``data`` to affect the
            determinism of the workflow, you can provide the function to compute the
            unique id of ``data`` using ``data_determiner``
        """
        if isinstance(data, WorkflowDataFrame):
            assert_or_throw(
                data.workflow is self,
                lambda: FugueWorkflowCompileError(
                    f"{data} does not belong to this workflow"
                ),
            )
            assert_or_throw(
                schema is None,
                FugueWorkflowCompileError(
                    "schema must be None when data is WorkflowDataFrame"
                ),
            )
            self._last_df = data
            return data
        if (
            (isinstance(data, (List, Iterable)) and not isinstance(data, str))
            or isinstance(data, Yielded)
            or is_df(data)
        ):
            return self.create(
                using=CreateData(
                    data,
                    schema=schema,
                    data_determiner=data_determiner,
                )
            )
        raise FugueWorkflowCompileError(
            f"Input data of type {type(data)} can't "
            "be converted to WorkflowDataFrame"
        )

    def df(
        self,
        data: Any,
        schema: Any = None,
        data_determiner: Optional[Callable[[Any], str]] = None,
    ) -> WorkflowDataFrame:
        """Create dataframe. Alias of :meth:`~.create_data`

        :param data: |DataFrameLikeObject| or :class:`~fugue.workflow.yielded.Yielded`
        :param schema: |SchemaLikeObject|, defaults to None
        :param data_determiner: a function to compute unique id from ``data``
        :return: a dataframe of the current workflow

        .. note::

            By default, the input ``data`` does not affect the determinism of the
            workflow (but ``schema`` and ``etadata`` do), because the amount of
            compute can be unpredictable. But if you want ``data`` to affect the
            determinism of the workflow, you can provide the function to compute
            the unique id of ``data`` using ``data_determiner``
        """
        return self.create_data(
            data=data, schema=schema, data_determiner=data_determiner
        )

    def load(
        self, path: str, fmt: str = "", columns: Any = None, **kwargs: Any
    ) -> WorkflowDataFrame:
        """Load dataframe from persistent storage.
        Read :ref:`this <tutorial:tutorials/advanced/dag:save & load>`
        for details.

        :param path: file path
        :param fmt: format hint can accept ``parquet``, ``csv``, ``json``,
          defaults to "", meaning to infer
        :param columns: list of columns or a |SchemaLikeObject|, defaults to None
        :return: dataframe from the file
        :rtype: WorkflowDataFrame
        """
        return self.create(
            using=Load, params=dict(path=path, fmt=fmt, columns=columns, params=kwargs)
        )

    def show(
        self,
        *dfs: Any,
        n: int = 10,
        with_count: bool = False,
        title: Optional[str] = None,
    ) -> None:
        """Show the dataframes.
        See
        :ref:`examples <tutorial:tutorials/advanced/dag:initialize a workflow>`.

        :param dfs: |DataFramesLikeObject|
        :param n: max number of rows, defaults to 10
        :param with_count: whether to show total count, defaults to False
        :param title: title to display on top of the dataframe, defaults to None
        :param best_width: max width for the output table, defaults to 100

        .. note::

            * When you call this method, it means you want the dataframe to be
              printed when the workflow executes. So the dataframe won't show until
              you run the workflow.
            * When ``with_count`` is True, it can trigger expensive calculation for
              a distributed dataframe. So if you call this function directly, you may
              need to :meth:`~.WorkflowDataFrame.persist` the dataframe. Or you can
              turn on |AutoPersist|
        """
        self.output(
            *dfs, using=Show, params=dict(n=n, with_count=with_count, title=title)
        )

    def join(
        self, *dfs: Any, how: str, on: Optional[Iterable[str]] = None
    ) -> WorkflowDataFrame:
        """Join dataframes.
        |ReadJoin|

        :param dfs: |DataFramesLikeObject|
        :param how: can accept ``semi``, ``left_semi``, ``anti``, ``left_anti``,
          ``inner``, ``left_outer``, ``right_outer``, ``full_outer``, ``cross``
        :param on: it can always be inferred, but if you provide, it will be
          validated against the inferred keys. Default to None
        :return: joined dataframe
        """
        _on: List[str] = list(on) if on is not None else []
        return self.process(*dfs, using=RunJoin, params=dict(how=how, on=_on))

    def set_op(self, how: str, *dfs: Any, distinct: bool = True) -> WorkflowDataFrame:
        """Union, subtract or intersect dataframes.

        :param how: can accept ``union``, ``left_semi``, ``anti``, ``left_anti``,
          ``inner``, ``left_outer``, ``right_outer``, ``full_outer``, ``cross``
        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after the set operation,
          default to True
        :return: result dataframe of the set operation

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        return self.process(
            *dfs, using=RunSetOperation, params=dict(how=how, distinct=distinct)
        )

    def union(self, *dfs: Any, distinct: bool = True) -> WorkflowDataFrame:
        """Union dataframes in ``dfs``.

        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after union,
          default to True
        :return: unioned dataframe

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        return self.set_op("union", *dfs, distinct=distinct)

    def subtract(self, *dfs: Any, distinct: bool = True) -> WorkflowDataFrame:
        """Subtract ``dfs[1:]`` from ``dfs[0]``.

        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after subtraction,
          default to True
        :return: subtracted dataframe

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        return self.set_op("subtract", *dfs, distinct=distinct)

    def intersect(self, *dfs: Any, distinct: bool = True) -> WorkflowDataFrame:
        """Intersect dataframes in ``dfs``.

        :param dfs: |DataFramesLikeObject|
        :param distinct: whether to perform `distinct` after intersection,
          default to True
        :return: intersected dataframe

        .. note::

            Currently, all dataframes in ``dfs`` must have identical schema, otherwise
            exception will be thrown.
        """
        return self.set_op("intersect", *dfs, distinct=distinct)

    def zip(
        self,
        *dfs: Any,
        how: str = "inner",
        partition: Any = None,
        temp_path: Optional[str] = None,
        to_file_threshold: Any = -1,
    ) -> WorkflowDataFrame:
        """Zip multiple dataframes together with given partition
        specifications.

        :param dfs: |DataFramesLikeObject|
        :param how: can accept ``inner``, ``left_outer``, ``right_outer``,
          ``full_outer``, ``cross``, defaults to ``inner``
        :param partition: |PartitionLikeObject|, defaults to None.
        :param temp_path: file path to store the data (used only if the serialized data
          is larger than ``to_file_threshold``), defaults to None
        :param to_file_threshold: file byte size threshold, defaults to -1

        :return: a zipped dataframe

        .. note::

            * If ``dfs`` is dict like, the zipped dataframe will be dict like,
              If ``dfs`` is list like, the zipped dataframe will be list like
            * It's fine to contain only one dataframe in ``dfs``

        .. seealso::

            Read |CoTransformer| and |ZipComap| for details
        """
        return self.process(
            *dfs,
            using=Zip,
            params=dict(
                how=how, temp_path=temp_path, to_file_threshold=to_file_threshold
            ),
            pre_partition=partition,
        )

    def transform(
        self,
        *dfs: Any,
        using: Any,
        schema: Any = None,
        params: Any = None,
        pre_partition: Any = None,
        ignore_errors: List[Any] = _DEFAULT_IGNORE_ERRORS,
        callback: Any = None,
    ) -> WorkflowDataFrame:
        """Transform dataframes using transformer.

        Please read |TransformerTutorial|

        :param dfs: |DataFramesLikeObject|
        :param using: transformer-like object, if it is a string, then it must be
          the alias of a registered transformer/cotransformer
        :param schema: |SchemaLikeObject|, defaults to None. The transformer
          will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
        :param params: |ParamsLikeObject| to run the processor, defaults to None.
          The transformer will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None. It's
          recommended to use the equivalent wayt, which is to call
          :meth:`~.partition` and then call :meth:`~.transform` without this parameter
        :param ignore_errors: list of exception types the transformer can ignore,
          defaults to empty list
        :param callback: |RPCHandlerLikeObject|, defaults to None
        :return: the transformed dataframe

        .. note::

            :meth:`~.transform` can be lazy and will return the transformed dataframe,
            :meth:`~.out_transform` is guaranteed to execute immediately (eager) and
            return nothing
        """
        assert_or_throw(
            len(dfs) == 1,
            NotImplementedError("transform supports only single dataframe"),
        )
        tf = _to_transformer(using, schema)
        tf._partition_spec = PartitionSpec(pre_partition)  # type: ignore
        callback = to_rpc_handler(callback)
        tf._has_rpc_client = not isinstance(callback, EmptyRPCHandler)  # type: ignore
        tf.validate_on_compile()
        return self.process(
            *dfs,
            using=RunTransformer,
            schema=None,
            params=dict(
                transformer=tf,
                ignore_errors=ignore_errors,
                params=params,
                rpc_handler=callback,
            ),
            pre_partition=pre_partition,
        )

    def out_transform(
        self,
        *dfs: Any,
        using: Any,
        params: Any = None,
        pre_partition: Any = None,
        ignore_errors: List[Any] = _DEFAULT_IGNORE_ERRORS,
        callback: Any = None,
    ) -> None:
        """Transform dataframes using transformer, it materializes the execution
        immediately and returns nothing

        Please read |TransformerTutorial|

        :param dfs: |DataFramesLikeObject|
        :param using: transformer-like object, if it is a string, then it must be
          the alias of a registered output transformer/cotransformer
        :param schema: |SchemaLikeObject|, defaults to None. The transformer
          will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
        :param params: |ParamsLikeObject| to run the processor, defaults to None.
          The transformer will be able to access this value from
          :meth:`~fugue.extensions.context.ExtensionContext.params`
        :param pre_partition: |PartitionLikeObject|, defaults to None. It's
          recommended to use the equivalent wayt, which is to call
          :meth:`~.partition` and then call :meth:`~.out_transform` without this
          parameter
        :param ignore_errors: list of exception types the transformer can ignore,
          defaults to empty list
        :param callback: |RPCHandlerLikeObject|, defaults to None

        .. note::

            :meth:`~.transform` can be lazy and will return the transformed dataframe,
            :meth:`~.out_transform` is guaranteed to execute immediately (eager) and
            return nothing
        """
        assert_or_throw(
            len(dfs) == 1,
            NotImplementedError("output transform supports only single dataframe"),
        )
        tf = _to_output_transformer(using)
        tf._partition_spec = PartitionSpec(pre_partition)  # type: ignore
        callback = to_rpc_handler(callback)
        tf._has_rpc_client = not isinstance(callback, EmptyRPCHandler)  # type: ignore
        tf.validate_on_compile()
        self.output(
            *dfs,
            using=RunOutputTransformer,
            params=dict(
                transformer=tf,
                ignore_errors=ignore_errors,
                params=params,
                rpc_handler=callback,
            ),
            pre_partition=pre_partition,
        )

    def select(
        self,
        *statements: Any,
        sql_engine: Any = None,
        sql_engine_params: Any = None,
        dialect: Optional[str] = FUGUE_SQL_DEFAULT_DIALECT,
    ) -> WorkflowDataFrame:
        """Execute ``SELECT`` statement using
        :class:`~fugue.execution.execution_engine.SQLEngine`

        :param statements: a list of sub-statements in string
          or :class:`~.WorkflowDataFrame`
        :param sql_engine: it can be empty string or null (use the default SQL
          engine), a string (use the registered SQL engine), an
          :class:`~fugue.execution.execution_engine.SQLEngine` type, or
          the :class:`~fugue.execution.execution_engine.SQLEngine` instance
          (you can use ``None`` to use the default one), defaults to None
        :return: result of the ``SELECT`` statement

        .. admonition:: Examples

            .. code-block:: python

                with FugueWorkflow() as dag:
                    a = dag.df([[0,"a"]],a:int,b:str)
                    b = dag.df([[0]],a:int)
                    c = dag.select("SELECT a FROM",a,"UNION SELECT * FROM",b)
                dag.run()

        Please read :ref:`this <tutorial:tutorials/advanced/dag:select query>`
        for more examples
        """
        sql: List[Tuple[bool, str]] = []
        dfs: Dict[str, DataFrame] = {}
        for s in statements:
            if isinstance(s, str):
                sql.append((False, s))
            else:
                ws = self.df(s)
                dfs[ws.name] = ws
                sql.append((True, ws.name))
        if sql[0][0]:  # starts with reference
            sql.insert(0, (False, "SELECT"))
        else:  # start with string but without select
            start = sql[0][1].strip()
            if not start[:10].upper().startswith("SELECT") and not start[
                :10
            ].upper().startswith("WITH"):
                sql[0] = (False, "SELECT " + start)
        return self.process(
            dfs,
            using=RunSQLSelect,
            params=dict(
                statement=StructuredRawSQL(sql, dialect=dialect),
                sql_engine=sql_engine,
                sql_engine_params=ParamDict(sql_engine_params),
            ),
        )

    def assert_eq(self, *dfs: Any, **params: Any) -> None:
        """Compare if these dataframes are equal. It's for internal, unit test
        purpose only. It will convert both dataframes to
        :class:`~fugue.dataframe.dataframe.LocalBoundedDataFrame`, so it assumes
        all dataframes are small and fast enough to convert. DO NOT use it
        on critical or expensive tasks.

        :param dfs: |DataFramesLikeObject|
        :param digits: precision on float number comparison, defaults to 8
        :param check_order: if to compare the row orders, defaults to False
        :param check_schema: if compare schemas, defaults to True
        :param check_content: if to compare the row values, defaults to True
        :param no_pandas: if true, it will compare the string representations of the
          dataframes, otherwise, it will convert both to pandas dataframe to compare,
          defaults to False

        :raises AssertionError: if not equal
        """
        self.output(*dfs, using=AssertEqual, params=params)

    def assert_not_eq(self, *dfs: Any, **params: Any) -> None:
        """Assert if all dataframes are not equal to the first dataframe.
        It's for internal, unit test purpose only. It will convert both dataframes to
        :class:`~fugue.dataframe.dataframe.LocalBoundedDataFrame`, so it assumes
        all dataframes are small and fast enough to convert. DO NOT use it
        on critical or expensive tasks.

        :param dfs: |DataFramesLikeObject|
        :param digits: precision on float number comparison, defaults to 8
        :param check_order: if to compare the row orders, defaults to False
        :param check_schema: if compare schemas, defaults to True
        :param check_content: if to compare the row values, defaults to True
        :param no_pandas: if true, it will compare the string representations of the
          dataframes, otherwise, it will convert both to pandas dataframe to compare,
          defaults to False

        :raises AssertionError: if any dataframe equals to the first dataframe
        """
        self.output(*dfs, using=AssertNotEqual, params=params)

    def add(self, task: FugueTask, *args: Any, **kwargs: Any) -> WorkflowDataFrame:
        """This method should not be called directly by users. Use
        :meth:`~.create`, :meth:`~.process`, :meth:`~.output` instead
        """
        if self.conf.get_or_throw(
            FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE, bool
        ) and sys.version_info >= (3, 7):
            dep = self.conf.get_or_throw(FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT, int)
            if dep > 0:
                conf = self.conf.get_or_throw(FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE, str)
                pre = [p for p in conf.split(",") if p != ""]
                task.reset_traceback(
                    limit=dep,
                    should_prune=lambda x: any(x.lower().startswith(xx) for xx in pre),
                )
        assert_or_throw(task._node_spec is None, lambda: f"can't reuse {task}")
        dep = _Dependencies(self, task, {}, *args, **kwargs)
        name = "_" + str(len(self._spec.tasks))
        wt = self._spec.add_task(name, task, dep.dependency)
        # TODO: this is auto persist, the implementation needs imrpovement
        for v in dep.dependency.values():
            v = v.split(".")[0]
            self._graph.add(name, v)
            if len(self._graph.down[v]) > 1 and self.conf.get_or_throw(
                FUGUE_CONF_WORKFLOW_AUTO_PERSIST, bool
            ):
                self._spec.tasks[v].set_checkpoint(
                    WeakCheckpoint(
                        lazy=False,
                        level=self.conf.get_or_none(
                            FUGUE_CONF_WORKFLOW_AUTO_PERSIST_VALUE, object
                        ),
                    )
                )

        return WorkflowDataFrame(self, wt)

    def _to_dfs(self, *args: Any, **kwargs: Any) -> DataFrames:
        return DataFrames(*args, **kwargs).convert(self.create_data)

    def __getattr__(self, name: str) -> Any:  # pragma: no cover
        """The dummy method to avoid PyLint complaint"""
        raise AttributeError(name)


class _Dependencies:
    def __init__(
        self,
        workflow: "FugueWorkflow",
        task: FugueTask,
        local_vars: Dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ):
        self.workflow = workflow
        self._local_vars = local_vars
        self.dependency: Dict[str, str] = {}
        for i in range(len(args)):
            key = task.inputs.get_key_by_index(i)
            self.dependency[key] = self._parse_single_dependency(args[i])
        for k, v in kwargs.items():
            self.dependency[k] = self._parse_single_dependency(v)

    def _parse_single_dependency(self, dep: Any) -> str:
        # if isinstance(dep, tuple):  # (cursor_like_obj, output_name)
        #     cursor = self._parse_cursor(dep[0])
        #     return cursor._task.name + "." + dep[1]
        return self._parse_cursor(dep)._task.single_output_expression

    def _parse_cursor(self, dep: Any) -> WorkflowDataFrame:
        if isinstance(dep, WorkflowDataFrame):
            return dep
        # if isinstance(dep, DataFrame):
        #     return self.workflow.create_data(dep)
        # if isinstance(dep, str):
        #     assert_or_throw(
        #         dep in self._local_vars, KeyError(f"{dep} is not a local variable")
        #     )
        #     if isinstance(self._local_vars[dep], WorkflowDataFrame):
        #         return self._local_vars[dep]
        #     # TODO: should also accept dataframe?
        #     raise TypeError(f"{self._local_vars[dep]} is not a valid dependency type")
        raise TypeError(f"{dep} is not a valid dependency type")  # pragma: no cover


# TODO: this should not exist, dependency libraries should do the job
class _Graph:
    def __init__(self):
        self.down: Dict[str, Set[str]] = defaultdict(set)
        self.up: Dict[str, Set[str]] = defaultdict(set)

    def add(self, name: str, depend_on: str) -> None:
        depend_on = depend_on.split(".")[0]
        self.down[depend_on].add(name)
        self.up[name].add(depend_on)



import os

from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.pandas_dataframe import PandasDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue._utils.io import FileParser, load_df, save_df, _FORMAT_MAP
from fugue.exceptions import FugueDataFrameOperationError
from pytest import raises
from triad.collections.fs import FileSystem
from triad.exceptions import InvalidOperationError


def test_file_parser():
    f = FileParser("c.parquet")
    assert "c.parquet" == f.uri
    assert "c.parquet" == f.uri_with_glob
    assert "" == f.scheme
    assert "c.parquet" == f.path
    assert ".parquet" == f.suffix
    assert "parquet" == f.file_format
    assert "" == f.glob_pattern
    assert "." == f.parent

    f = FileParser("/a/b/c.parquet")
    assert "/a/b/c.parquet" == f.uri
    assert "/a/b/c.parquet" == f.uri_with_glob
    assert "" == f.scheme
    assert "/a/b/c.parquet" == f.path
    assert ".parquet" == f.suffix
    assert "parquet" == f.file_format
    assert "" == f.glob_pattern
    assert "/a/b" == f.parent

    for k, v in _FORMAT_MAP.items():
        f = FileParser(f"s3://a/b/c{k}")
        assert f"s3://a/b/c{k}" == f.uri
        assert "s3" == f.scheme
        assert f"/b/c{k}" == f.path
        assert k == f.suffix
        assert v == f.file_format
        assert "s3://a/b" == f.parent

    f = FileParser("s3://a/b/c.test.parquet")
    assert "s3://a/b/c.test.parquet" == f.uri
    assert "s3" == f.scheme
    assert "/b/c.test.parquet" == f.path
    assert ".test.parquet" == f.suffix
    assert "parquet" == f.file_format
    assert "s3://a/b" == f.parent

    f = FileParser("s3://a/b/c.ppp.gz", "csv")
    assert "s3://a/b/c.ppp.gz" == f.uri
    assert "s3" == f.scheme
    assert "/b/c.ppp.gz" == f.path
    assert ".ppp.gz" == f.suffix
    assert "csv" == f.file_format

    f = FileParser("s3://a/b/c", "csv")
    assert "s3://a/b/c" == f.uri
    assert "s3" == f.scheme
    assert "/b/c" == f.path
    assert "" == f.suffix
    assert "csv" == f.file_format

    raises(NotImplementedError, lambda: FileParser("s3://a/b/c.ppp"))
    raises(NotImplementedError, lambda: FileParser("s3://a/b/c.parquet", "csvv"))
    raises(NotImplementedError, lambda: FileParser("s3://a/b/c"))


def test_file_parser_glob():
    f = FileParser("/a/b/*.parquet")
    assert "/a/b" == f.uri
    assert "" == f.scheme
    assert "/a/b/*.parquet" == f.path
    assert ".parquet" == f.suffix
    assert "parquet" == f.file_format
    assert "*.parquet" == f.glob_pattern
    assert "/a/b/*.parquet" == f.uri_with_glob

    f = FileParser("/a/b/*123.parquet")
    assert "/a/b" == f.uri
    assert "" == f.scheme
    assert "/a/b/*123.parquet" == f.path
    assert ".parquet" == f.suffix
    assert "parquet" == f.file_format
    assert "*123.parquet" == f.glob_pattern
    assert "/a/b/*123.parquet" == f.uri_with_glob

    f = FileParser("s3://a/b/*.parquet")
    assert "s3://a/b" == f.uri
    assert "s3" == f.scheme
    assert "/b/*.parquet" == f.path
    assert ".parquet" == f.suffix
    assert "parquet" == f.file_format
    assert "*.parquet" == f.glob_pattern
    assert "s3://a/b/*.parquet" == f.uri_with_glob

    ff = FileParser("s3://a/b", "parquet").with_glob("*.csv", "csv")
    assert "s3://a/b/*.csv" == ff.uri_with_glob
    assert "csv" == ff.file_format
    ff = FileParser("s3://a/b/", "csv").with_glob("*.csv")
    assert "s3://a/b/*.csv" == ff.uri_with_glob
    assert "csv" == ff.file_format
    ff = FileParser("s3://a/b/*.parquet").with_glob("*.csv")
    assert "s3://a/b/*.csv" == ff.uri_with_glob
    assert "csv" == ff.file_format
    ff = FileParser("s3://a/b/*.parquet", "parquet").with_glob("*.csv")
    assert "s3://a/b/*.csv" == ff.uri_with_glob
    assert "parquet" == ff.file_format
    ff = FileParser("s3://a/b/*.parquet", "parquet").with_glob("*.csv", "csv")
    assert "s3://a/b/*.csv" == ff.uri_with_glob
    assert "csv" == ff.file_format


def test_parquet_io(tmpdir):
    df1 = PandasDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    df2 = ArrayDataFrame([[[1, 2]]], "a:[int]")
    # {a:int} will become {a:long} because pyarrow lib has issue
    df3 = ArrayDataFrame([[dict(a=1)]], "a:{a:long}")
    for df in [df1, df2, df3]:
        path = os.path.join(tmpdir, "a.parquet")
        save_df(df, path)
        actual = load_df(path)
        df_eq(df, actual, throw=True)

    save_df(df1, path)
    actual = load_df(path, columns=["b", "a"])
    df_eq(actual, [[2, "1"]], "b:int,a:str")
    actual = load_df(path, columns="b:str,a:int")
    df_eq(actual, [["2", 1]], "b:str,a:int")
    # can't specify wrong columns
    raises(Exception, lambda: load_df(path, columns="bb:str,a:int"))

    # load directory
    fs = FileSystem()
    for name in ["folder.parquet", "folder"]:
        folder = os.path.join(tmpdir, name)
        fs.makedirs(folder)
        f0 = os.path.join(folder, "_SUCCESS")
        f1 = os.path.join(folder, "1.parquet")
        f2 = os.path.join(folder, "3.parquet")
        fs.touch(f0)
        save_df(df1, f1)
        save_df(df1, f2)

    actual = load_df(folder, "parquet")
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # load multiple paths
    actual = load_df([f1, f2], "parquet")
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # load folder
    actual = load_df(folder, "parquet")
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    actual = load_df(os.path.join(tmpdir, "folder.parquet"))
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # load pattern
    actual = load_df(os.path.join(tmpdir, "folder", "*.parquet"))
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # overwrite folder with single file
    save_df(actual, os.path.join(tmpdir, "folder.parquet"), mode="overwrite")
    actual = load_df(os.path.join(tmpdir, "folder.parquet"))
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # overwrite = False
    raises(FileExistsError, lambda: save_df(df1, f1, mode="error"))
    raises(
        FileExistsError,
        lambda: save_df(df1, os.path.join(tmpdir, "folder.parquet"), mode="error"),
    )

    # wrong mode
    raises(NotImplementedError, lambda: save_df(df1, f1, mode="dummy"))


def test_csv_io(tmpdir):
    fs = FileSystem()
    df1 = PandasDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    path = os.path.join(tmpdir, "a.csv")
    # without header
    save_df(df1, path)
    assert fs.readtext(path).startswith("1,2,3")
    raises(ValueError, lambda: load_df(path, header=False))
    actual = load_df(path, columns=["a", "b", "c"], header=False, infer_schema=True)
    assert [[1, 2, 3]] == actual.as_array()
    assert actual.schema == "a:long,b:long,c:long"
    actual = load_df(path, columns="a:double,b:str,c:str", header=False)
    assert [[1.0, "2", "3"]] == actual.as_array()
    assert actual.schema == "a:double,b:str,c:str"
    # with header
    save_df(df1, path, header=True)
    assert fs.readtext(path).startswith("a,b,c")
    actual = load_df(path, header=True)
    assert [["1", "2", "3"]] == actual.as_array()
    actual = load_df(path, header=True, infer_schema=True)
    assert [[1, 2, 3]] == actual.as_array()
    actual = load_df(path, columns=["b", "a"], header=True, infer_schema=True)
    assert [[2, 1]] == actual.as_array()
    actual = load_df(path, columns="b:str,a:double", header=True)
    assert [["2", 1.0]] == actual.as_array()
    raises(KeyError, lambda: load_df(path, columns="b:str,x:double", header=True))

    raises(
        NotImplementedError, lambda: load_df(path, columns="b:str,x:double", header=2)
    )


def test_json(tmpdir):
    fs = FileSystem()
    df1 = PandasDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    path = os.path.join(tmpdir, "a.json")
    save_df(df1, path)
    actual = load_df(path)
    df_eq(actual, [[1, 2, 3]], "a:long,b:long,c:long")
    actual = load_df(path, columns=["b", "a"])
    df_eq(actual, [[2, "1"]], "b:int,a:str")
    actual = load_df(path, columns="b:str,a:int")
    df_eq(actual, [["2", 1]], "b:str,a:int")
    raises(KeyError, lambda: load_df(path, columns="bb:str,a:int"))


def test_avro_io(tmpdir):
    df1 = PandasDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    df2 = PandasDataFrame([["hello", 2, 3]], "a:str,b:int,c:long")
    path1 = os.path.join(tmpdir, "df1.avro")
    path2 = os.path.join(tmpdir, "df2.avro")
    save_df(df1, path1)
    actual = load_df(path1)

    df_eq(actual, [["1", 2, 3]], "a:str,b:long,c:long")
    actual = load_df(path1, columns=["a", "b"])
    df_eq(actual, [["1", 3]], "a:str,b:long")

    actual = load_df(path1, columns="a:str,b:int,c:long")
    df_eq(actual, [["1", 2, 3]], "a:str,b:int,c:long")

    actual = load_df(
        path1, columns="a:str,b:int,c:long", infer_schema=True
    )  # TODO raise error when both provided?
    df_eq(actual, [["1", 2, 3]], "a:str,b:int,c:long")

    actual = load_df(path1, columns=["b", "c"], infer_schema=True)
    df_eq(actual, [[2, 3]], "b:long,c:long")

    # save in append mode
    path3 = os.path.join(tmpdir, "append.avro")
    save_df(df1, path3)
    save_df(df2, path3, append=True)
    actual = load_df(path1, columns="a:str,b:int,c:long")
    df_eq(actual, [["1", 2, 3], ["hello", 2, 3]], "a:str,b:int,c:long")

    # save times_as_micros =False (i.e milliseconds instead)
    df4 = PandasDataFrame([["2021-05-04", 2, 3]], "a:datetime,b:int,c:long")
    path4 = os.path.join(tmpdir, "df4.avro")
    save_df(df4, path4)
    actual = load_df(path4, columns="a:datetime,b:int,c:long")
    df_eq(actual, [["2021-05-04", 2, 3]], "a:datetime,b:int,c:long")
    save_df(df4, path4, times_as_micros=False)
    actual = load_df(path4, columns="a:datetime,b:int,c:long")
    df_eq(actual, [["2021-05-04", 2, 3]], "a:datetime,b:int,c:long")

    # provide avro schema
    schema = {
        "type": "record",
        "name": "Root",
        "fields": [
            {"name": "a", "type": "string"},
            {"name": "b", "type": "int"},
            {"name": "c", "type": "long"},
        ],
    }
    save_df(df1, path1, schema=schema)
    actual = load_df(path1, columns="a:str,b:int,c:long")
    df_eq(actual, [["1", 2, 3]], "a:str,b:int,c:long")

    # provide wrong types in columns arg
    save_df(df2, path2, schema=schema)
    raises(
        FugueDataFrameOperationError,
        lambda: load_df(df2, path2, columns="a:int,b:int,c:long"),
    )

    # load with process_record function
    actual = load_df(
        path2,
        columns="a:str,b:int,c:long",
        process_record=lambda s: {"a": str.upper(s["a"]), "b": s["b"], "c": s["c"]},
    )
    df_eq(actual, [["HELLO", 2, 3]], "a:str,b:int,c:long")

    # provide wrong type in avro schema
    schema = {
        "type": "record",
        "name": "Root",
        "fields": [
            {"name": "a", "type": "int"},
            {"name": "b", "type": "int"},
            {"name": "c", "type": "long"},
        ],
    }
    raises(TypeError, lambda: save_df(df2, path2, schema=schema))



from fugue import FugueWorkflow


def test_importless():
    for engine in [None]:
        dag = FugueWorkflow()
        idf = dag.df([[0], [1]], "a:int").as_ibis()
        idf[idf.a < 1].as_fugue().show()

        dag.run(engine)



import copy

import pandas as pd
from pytest import raises
from triad.collections.schema import Schema

from fugue.dataframe import ArrayDataFrame, DataFrame
from fugue.api import as_fugue_df, get_native_as_df
from fugue.bag.array_bag import ArrayBag


def test_as_fugue_df():
    with raises(NotImplementedError):
        as_fugue_df(10)
    with raises(TypeError):
        as_fugue_df(ArrayBag([1, 2]))
    df = pd.DataFrame([[0]], columns=["a"])
    assert isinstance(as_fugue_df(df), DataFrame)


def test_get_native_as_df():
    with raises(NotImplementedError):
        get_native_as_df(10)
    # other tests are in the suites


def test_show():
    df = ArrayDataFrame(schema="a:str,b:str")
    df.show()

    assert repr(df) == df._repr_html_()

    s = " ".join(["x"] * 2)
    df = ArrayDataFrame([[s, 1], ["b", 2]], "a:str,b:str")
    df.show()

    s = " ".join(["x"] * 200)
    df = ArrayDataFrame([[s, 1], ["b", 2]], "a:str,b:str")
    df.show()

    s = " ".join(["x"] * 200)
    df = ArrayDataFrame([[s, 1], ["b", s]], "a:str,b:str")
    df.show()

    s = "".join(["x"] * 2000)
    df = ArrayDataFrame([[s, 1], ["b", None]], "a:str,b:str")
    df.show()

    s = " ".join(["x"] * 20)
    schema = [f"a{x}:str" for x in range(20)]
    data = [[f"aasdfjasdfka;sdf{x}:str" for x in range(20)]]
    df = ArrayDataFrame(data, schema)
    df.show()

    s = " ".join(["x"] * 200)
    df = ArrayDataFrame([[s, 1], ["b", "s"]], "a:str,b:str")
    df.show(n=1, with_count=True, title="abc")


def test_lazy_schema():
    df = MockDF([["a", 1], ["b", 2]], "a:str,b:str")
    assert callable(df._schema)
    assert df.schema == "a:str,b:str"


def test_get_info_str():
    df = ArrayDataFrame([["a", 1], ["b", 2]], "a:str,b:str")
    assert '{"schema": "a:str,b:str", "type": '
    '"tests.collections.dataframe.test_dataframe.MockDF", "metadata": {}}' == df.get_info_str()


def test_copy():
    df = ArrayDataFrame([["a", 1], ["b", 2]], "a:str,b:str")
    assert copy.copy(df) is df
    assert copy.deepcopy(df) is df


class MockDF(ArrayDataFrame):
    def __init__(self, df=None, schema=None):
        super().__init__(df=df, schema=schema)
        DataFrame.__init__(self, lambda: Schema(schema))



import pytest

ibis = pytest.importorskip("ibis")
import ibis.expr.datatypes as dt
import pandas as pd
import pyarrow as pa
from pytest import raises
from triad import Schema

from fugue import PandasDataFrame
from fugue.dataframe.utils import _df_eq
from fugue_ibis._utils import LazyIbisObject, materialize, to_ibis_schema, to_schema


def test_schema():
    a = Schema(
        "a:bool,b:int8,c:uint8,d:int16,e:uint16,f:int32,g:uint32,h:int64,i:uint64"
    )
    b = ibis.schema(
        [
            ("a", "boolean"),
            ("b", "int8"),
            ("c", "uint8"),
            ("d", "int16"),
            ("e", "uint16"),
            ("f", "int32"),
            ("g", "uint32"),
            ("h", "int64"),
            ("i", "uint64"),
        ]
    )
    assert to_ibis_schema(a) == b
    assert a == to_schema(b)

    a = Schema("a:float32,b:float64,c:datetime,d:date,e:binary,f:string")
    b = ibis.schema(
        [
            ("a", "float32"),
            ("b", "float64"),
            ("c", "timestamp"),
            ("d", "date"),
            ("e", "binary"),
            ("f", "string"),
        ]
    )
    assert to_ibis_schema(a) == b
    assert a == to_schema(b)

    a = Schema("a:decimal(10,2)")
    assert to_schema(to_ibis_schema(a)) == a

    a = Schema("a:[int],b:[{a:str}],c:{a:str},d:{a:[int]}")
    assert to_schema(to_ibis_schema(a)) == a

    a = Schema("a:<int,str>")
    assert to_schema(to_ibis_schema(a)) == a

    a = Schema("a:timestamp(us, GMT)")
    assert to_schema(to_ibis_schema(a)) == a


def test_schema_incompatible():
    orig = ibis.schema([("x y", dt.GeoSpatial()), ("y", dt.String())])
    with raises(NotImplementedError):
        to_schema(orig)
    assert to_schema(orig, lambda n, t: pa.string()) == "`x y`:str,y:str"


def test_materialize():
    tdf1 = pd.DataFrame([[0, 1], [3, 4]], columns=["a", "b"])

    _test_expr(lambda a: a, a=tdf1)
    _test_expr(lambda a: a[a.b], a=tdf1)
    _test_expr(lambda a: a[a.a >= 1], a=tdf1)
    _test_expr(lambda a: a[a.a.isin([0, 2])], a=tdf1)
    _test_expr(lambda a: a[a.b, a.a], a=tdf1)
    _test_expr(lambda a: a.b + ibis.literal(1, "int8"), a=tdf1)
    _test_expr(lambda a: 1 + a.b, a=tdf1)
    _test_expr(lambda a: a[(1 * a.a).name("x"), (1 / a.b).name("y")], a=tdf1)
    _test_expr(lambda a: a.group_by("a").aggregate(a.b.sum()), a=tdf1)

    tdf2 = pd.DataFrame([[0, "x"], [4, "y"]], columns=["a", "c"])
    _test_expr(lambda t, t2: t.join(t2, t.a == t2.a)[t, t2.c], t=tdf1, t2=tdf2)

    tdf3 = pd.DataFrame([[0, 1, 2], [3, 4, 5]], columns=["a", "b", "c"])
    _test_expr(
        lambda t: (
            t.group_by(["a", "b"])
            .aggregate(the_sum=t.c.sum())
            .group_by("a")
            .aggregate(mad=lambda x: x.the_sum.abs().mean())
        ),
        t=tdf3,
    )


def _test_expr(func, **dfs):
    con = ibis.pandas.connect(dfs)
    kwargs = {k: con.table(k) for k in dfs.keys()}
    expected = func(**kwargs).execute()

    lkwargs = {k: LazyIbisObject(k) for k in dfs.keys()}
    expr = func(**lkwargs)
    actual = materialize(expr, lambda k: kwargs[k]).execute()

    _df_eq(PandasDataFrame(actual), PandasDataFrame(expected), throw=True)



from fugue._utils.misc import get_attribute
from pytest import raises


def test_get_attribute():
    class C(object):
        pass

    c = C()
    assert "x" not in c.__dict__
    assert 0 == get_attribute(c, "x", int)
    assert 0 == c.x
    assert 0 == get_attribute(c, "x", int)
    c.x = 10
    assert 10 == get_attribute(c, "x", int)
    raises(TypeError, lambda: get_attribute(c, "x", str))



import pytest

ibis = pytest.importorskip("ibis")
from pyspark.sql import SparkSession

from fugue_ibis import IbisEngine
from fugue_spark import SparkExecutionEngine
from fugue_spark.ibis_engine import SparkIbisEngine
from fugue_test.ibis_suite import IbisTests


class SparkIbisTests(IbisTests.Tests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def make_engine(self):
        session = SparkSession.builder.getOrCreate()
        e = SparkExecutionEngine(session, dict(test=True))
        return e

    def make_ibis_engine(self) -> IbisEngine:
        return SparkIbisEngine(self._engine)



import json
import math
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
from pytest import raises
from triad.collections.schema import Schema

import fugue.api as fa
from fugue.dataframe import ArrowDataFrame, PandasDataFrame
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue_test.dataframe_suite import DataFrameTests


class PandasDataFrameTests(DataFrameTests.Tests):
    def df(self, data: Any = None, schema: Any = None) -> PandasDataFrame:
        return PandasDataFrame(data, schema)

    def test_num_partitions(self):
        assert fa.get_num_partitions(self.df([[0, 1]], "a:int,b:int")) == 1

    def test_api_as_local(self):
        assert fa.is_local(self.df([[0, 1]], "a:int,b:int"))


class NativePandasDataFrameTests(DataFrameTests.NativeTests):
    def df(self, data: Any = None, schema: Any = None) -> pd.DataFrame:
        return ArrowDataFrame(data, schema).as_pandas()

    def to_native_df(self, pdf: pd.DataFrame) -> Any:  # pragma: no cover
        return pdf

    def test_num_partitions(self):
        assert fa.get_num_partitions(self.df([[0, 1]], "a:int,b:int")) == 1

    def test_map_type(self):
        pass


def test_init():
    df = PandasDataFrame(schema="a:str,b:int")
    assert df.is_bounded
    assert df.count() == 0
    assert df.schema == "a:str,b:int"
    assert Schema(df.native) == "a:str,b:int"

    pdf = pd.DataFrame([["a", 1], ["b", 2]])
    raises(Exception, lambda: PandasDataFrame(pdf))
    df = PandasDataFrame(pdf, "a:str,b:str")
    assert [["a", "1"], ["b", "2"]] == df.native.values.tolist()
    df = PandasDataFrame(pdf, "a:str,b:int")
    assert [["a", 1], ["b", 2]] == df.native.values.tolist()
    df = PandasDataFrame(pdf, "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.native.values.tolist()

    pdf = pd.DataFrame([["a", 1], ["b", 2]], columns=["a", "b"])["b"]
    assert isinstance(pdf, pd.Series)
    df = PandasDataFrame(pdf, "b:str")
    assert [["1"], ["2"]] == df.native.values.tolist()
    df = PandasDataFrame(pdf, "b:double")
    assert [[1.0], [2.0]] == df.native.values.tolist()

    pdf = pd.DataFrame([["a", 1], ["b", 2]], columns=["x", "y"])
    df = PandasDataFrame(pdf)
    assert df.schema == "x:str,y:long"
    df = PandasDataFrame(pdf, "y:str,x:str")
    assert [["1", "a"], ["2", "b"]] == df.native.values.tolist()
    ddf = PandasDataFrame(df)
    assert [["1", "a"], ["2", "b"]] == ddf.native.values.tolist()
    assert df.native is ddf.native  # no real copy happened

    df = PandasDataFrame([["a", 1], ["b", "2"]], "x:str,y:double")
    assert [["a", 1.0], ["b", 2.0]] == df.native.values.tolist()

    df = PandasDataFrame([], "x:str,y:double")
    assert [] == df.native.values.tolist()

    raises(Exception, lambda: PandasDataFrame(123))


def test_simple_methods():
    df = PandasDataFrame([], "a:str,b:int")
    assert df.as_pandas() is df.native
    assert df.empty
    assert 0 == df.count()
    assert df.is_local

    df = PandasDataFrame([["a", 1], ["b", "2"]], "x:str,y:double")
    assert df.as_pandas() is df.native
    assert not df.empty
    assert 2 == df.count()
    assert ["a", 1.0] == df.peek_array()
    assert dict(x="a", y=1.0) == df.peek_dict()


def test_nested():
    # data = [[dict(a=1, b=[3, 4], d=1.0)], [json.dumps(dict(b=[30, "40"]))]]
    # df = PandasDataFrame(data, "a:{a:str,b:[int]}")
    # a = df.as_array(type_safe=True)
    # assert [[dict(a="1", b=[3, 4])], [dict(a=None, b=[30, 40])]] == a

    data = [[[json.dumps(dict(b=[30, "40"]))]]]
    df = PandasDataFrame(data, "a:[{a:str,b:[int]}]")
    a = df.as_array(type_safe=True)
    assert [[[dict(a=None, b=[30, 40])]]] == a


def test_rename():
    df = PandasDataFrame([["a", 1]], "a:str,b:int")
    df2 = df.rename(columns=dict(a="aa"))
    assert isinstance(df, PandasDataFrame)
    df_eq(df2, [["a", 1]], "aa:str,b:int", throw=True)
    df_eq(df, [["a", 1]], "a:str,b:int", throw=True)


def test_as_array():
    df = PandasDataFrame([], "a:str,b:int")
    assert [] == df.as_array()
    assert [] == df.as_array(type_safe=True)
    assert [] == list(df.as_array_iterable())
    assert [] == list(df.as_array_iterable(type_safe=True))

    df = PandasDataFrame([["a", 1]], "a:str,b:int")
    assert [["a", 1]] == df.as_array()
    assert [["a", 1]] == df.as_array(["a", "b"])
    assert [[1, "a"]] == df.as_array(["b", "a"])

    # prevent pandas auto type casting
    df = PandasDataFrame([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)
    assert [[1.0, 1]] == df.as_array(["a", "b"])
    assert [[1, 1.0]] == df.as_array(["b", "a"])

    df = PandasDataFrame([[np.float64(1.0), 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)

    df = PandasDataFrame([[pd.Timestamp("2020-01-01"), 1.1]], "a:datetime,b:int")
    # df.native["a"] = pd.to_datetime(df.native["a"])
    assert [[datetime(2020, 1, 1), 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], datetime)
    # assert not isinstance(df.as_array()[0][0], pd.Timestamp)
    assert isinstance(df.as_array()[0][1], int)

    df = PandasDataFrame([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array(type_safe=True)
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)


def test_as_dict_iterable():
    df = PandasDataFrame([[pd.NaT, 1.1]], "a:datetime,b:int")
    assert [dict(a=None, b=1)] == list(df.as_dict_iterable())
    df = PandasDataFrame([["2020-01-01", 1.1]], "a:datetime,b:int")
    assert [dict(a=datetime(2020, 1, 1), b=1)] == list(df.as_dict_iterable())


def test_nan_none():
    df = ArrayDataFrame([[None, None]], "b:str,c:double")
    assert df.as_pandas().iloc[0, 0] is None
    arr = PandasDataFrame(df.as_pandas(), df.schema).as_array()[0]
    assert arr[0] is None
    assert math.isnan(arr[1])

    df = ArrayDataFrame([[None, None]], "b:int,c:bool")
    arr = PandasDataFrame(df.as_pandas(), df.schema).as_array(type_safe=True)[0]
    assert arr[0] is None
    assert arr[1] is None

    df = ArrayDataFrame([["a", 1.1], [None, None]], "b:str,c:double")
    arr = PandasDataFrame(df.as_pandas(), df.schema).as_array(type_safe=True)[1]
    assert arr[0] is None
    assert arr[1] is None


def _test_as_array_perf():
    s = Schema()
    arr = []
    for i in range(100):
        s.append(f"a{i}:int")
        arr.append(i)
    for i in range(100):
        s.append(f"b{i}:int")
        arr.append(float(i))
    for i in range(100):
        s.append(f"c{i}:str")
        arr.append(str(i))
    data = []
    for i in range(5000):
        data.append(list(arr))
    df = PandasDataFrame(data, s)
    res = df.as_array()
    res = df.as_array(type_safe=True)
    nts, ts = 0.0, 0.0
    for i in range(10):
        t = datetime.now()
        res = df.as_array()
        nts += (datetime.now() - t).total_seconds()
        t = datetime.now()
        res = df.as_array(type_safe=True)
        ts += (datetime.now() - t).total_seconds()
    print(nts, ts)



from fugue.rpc import make_rpc_server
from triad import ParamDict
import cloudpickle


def test_flask_service():
    # fugue.rpc.flask.FlaskRPCServer
    conf = ParamDict(
        {
            "fugue.rpc.server": "fugue.rpc.flask.FlaskRPCServer",
            "fugue.rpc.flask_server.host": "127.0.0.1",
            "fugue.rpc.flask_server.port": "1234",
            "fugue.rpc.flask_server.timeout": "2 sec",
        }
    )

    def k(value: str) -> str:
        return value + "x"

    def kk(a: int, b: int) -> int:
        return a + b

    def kkk(f: callable, a: int) -> int:
        return f(a)

    with make_rpc_server(conf).start() as server:
        assert "1234" == server.conf["fugue.rpc.flask_server.port"]
        with server.start():  # recursive start will take no effect
            client1 = cloudpickle.loads(cloudpickle.dumps(server.make_client(k)))
        assert "dddx" == client1("ddd")
        client2 = cloudpickle.loads(cloudpickle.dumps(server.make_client(kk)))
        assert 3 == client2(1, 2)
        assert "dddx" == client1("ddd")
        client3 = cloudpickle.loads(cloudpickle.dumps(server.make_client(kkk)))
        assert 3 == client3(lambda x: x + 1, 2)
        assert 3 == client2(1, 2)
        server.stop()  # extra stop in the end will take no effect



from typing import Any, Callable, Dict

import ibis
from fugue import DataFrame, DataFrames, Processor, WorkflowDataFrame
from fugue.exceptions import FugueWorkflowCompileError
from fugue.workflow.workflow import WorkflowDataFrames
from triad import assert_or_throw, extension_method

from ._utils import LazyIbisObject, _materialize
from .execution.ibis_engine import parse_ibis_engine

from ._compat import IbisTable


def run_ibis(
    ibis_func: Callable[[ibis.BaseBackend], IbisTable],
    ibis_engine: Any = None,
    **dfs: WorkflowDataFrame,
) -> WorkflowDataFrame:
    """Run an ibis workflow wrapped in ``ibis_func``

    :param ibis_func: the function taking in an ibis backend, and returning
      an Ibis TableExpr
    :param ibis_engine: an object that together with |ExecutionEngine|
      can determine :class:`~fugue_ibis.execution.ibis_engine.IbisEngine`
      , defaults to None
    :param dfs: dataframes in the same workflow
    :return: the output workflow dataframe

    .. admonition:: Examples

        .. code-block:: python

            import fugue as FugueWorkflow
            from fugue_ibis import run_ibis

            def func(backend):
                t = backend.table("tb")
                return t.mutate(b=t.a+1)

            dag = FugueWorkflow()
            df = dag.df([[0]], "a:int")
            result = run_ibis(func, tb=df)
            result.show()
    """
    wdfs = WorkflowDataFrames(**dfs)
    return wdfs.workflow.process(
        wdfs,
        using=_IbisProcessor,
        params=dict(ibis_func=ibis_func, ibis_engine=ibis_engine),
    )


@extension_method
def as_ibis(df: WorkflowDataFrame) -> IbisTable:
    """Convert the Fugue workflow dataframe to an ibis table for ibis
    operations.

    :param df: the Fugue workflow dataframe
    :return: the object representing the ibis table

    .. admonition:: Examples

        .. code-block:: python

            # non-magical approach
            import fugue as FugueWorkflow
            from fugue_ibis import as_ibis, as_fugue

            dag = FugueWorkflow()
            df1 = dag.df([[0]], "a:int")
            df2 = dag.df([[1]], "a:int")
            idf1 = as_ibis(df1)
            idf2 = as_ibis(df2)
            idf3 = idf1.union(idf2)
            result = idf3.mutate(b=idf3.a+1)
            as_fugue(result).show()

        .. code-block:: python

            # magical approach
            import fugue as FugueWorkflow
            import fugue_ibis  # must import

            dag = FugueWorkflow()
            idf1 = dag.df([[0]], "a:int").as_ibis()
            idf2 = dag.df([[1]], "a:int").as_ibis()
            idf3 = idf1.union(idf2)
            result = idf3.mutate(b=idf3.a+1).as_fugue()
            result.show()

    .. note::

        The magic is that when importing ``fugue_ibis``, the functions
        ``as_ibis`` and ``as_fugue`` are added to the correspondent classes
        so you can use them as if they are parts of the original classes.

        This is an idea similar to patching. Ibis uses this programming model
        a lot. Fugue provides this as an option.

    .. note::

        The returned object is not really a ``TableExpr``, it's a 'super lazy'
        object that will be translated into ``TableExpr`` at run time.
        This is because to compile an ibis execution graph, the input schemas
        must be known. However, in Fugue, this is not always true. For example
        if the previous step is to pivot a table, then the output schema can be
        known at runtime. So in order to be a part of Fugue, we need to be able to
        construct ibis expressions before knowing the input schemas.
    """
    return LazyIbisObject(df)  # type: ignore


@extension_method(class_type=LazyIbisObject)
def as_fugue(
    expr: IbisTable,
    ibis_engine: Any = None,
) -> WorkflowDataFrame:
    """Convert a lazy ibis object to Fugue workflow dataframe

    :param expr: the actual instance should be LazyIbisObject
    :return: the Fugue workflow dataframe

    .. admonition:: Examples

        .. code-block:: python

            # non-magical approach
            import fugue as FugueWorkflow
            from fugue_ibis import as_ibis, as_fugue

            dag = FugueWorkflow()
            df1 = dag.df([[0]], "a:int")
            df2 = dag.df([[1]], "a:int")
            idf1 = as_ibis(df1)
            idf2 = as_ibis(df2)
            idf3 = idf1.union(idf2)
            result = idf3.mutate(b=idf3.a+1)
            as_fugue(result).show()

        .. code-block:: python

            # magical approach
            import fugue as FugueWorkflow
            import fugue_ibis  # must import

            dag = FugueWorkflow()
            idf1 = dag.df([[0]], "a:int").as_ibis()
            idf2 = dag.df([[1]], "a:int").as_ibis()
            idf3 = idf1.union(idf2)
            result = idf3.mutate(b=idf3.a+1).as_fugue()
            result.show()

    .. note::

        The magic is that when importing ``fugue_ibis``, the functions
        ``as_ibis`` and ``as_fugue`` are added to the correspondent classes
        so you can use them as if they are parts of the original classes.

        This is an idea similar to patching. Ibis uses this programming model
        a lot. Fugue provides this as an option.

    .. note::

        The returned object is not really a ``TableExpr``, it's a 'super lazy'
        object that will be translated into ``TableExpr`` at run time.
        This is because to compile an ibis execution graph, the input schemas
        must be known. However, in Fugue, this is not always true. For example
        if the previous step is to pivot a table, then the output schema can be
        known at runtime. So in order to be a part of Fugue, we need to be able to
        construct ibis expressions before knowing the input schemas.
    """

    def _func(
        be: ibis.BaseBackend,
        lazy_expr: LazyIbisObject,
        ctx: Dict[int, Any],
    ) -> IbisTable:
        return _materialize(
            lazy_expr, {k: be.table(f"_{id(v)}") for k, v in ctx.items()}
        )

    assert_or_throw(
        isinstance(expr, LazyIbisObject),
        FugueWorkflowCompileError("expr must be a LazyIbisObject"),
    )
    _lazy_expr: LazyIbisObject = expr  # type: ignore
    _ctx = _lazy_expr._super_lazy_internal_ctx
    _dfs = {f"_{id(v)}": v for _, v in _ctx.items()}
    return run_ibis(
        lambda be: _func(be, _lazy_expr, _ctx), ibis_engine=ibis_engine, **_dfs
    )


class _IbisProcessor(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        ibis_func = self.params.get_or_throw("ibis_func", Callable)
        ibis_engine = self.params.get_or_none("ibis_engine", object)
        ie = parse_ibis_engine(
            self.execution_engine if ibis_engine is None else ibis_engine,
            self.execution_engine,
        )
        return ie.select(dfs, ibis_func)



from typing import Type, TypeVar

from triad.utils.assertion import assert_or_throw

T = TypeVar("T")


def get_attribute(obj: object, attr_name: str, data_type: Type[T]) -> T:
    if attr_name not in obj.__dict__ or obj.__dict__[attr_name] is None:
        obj.__dict__[attr_name] = data_type()
    assert_or_throw(
        isinstance(obj.__dict__[attr_name], data_type),
        lambda: TypeError(f"{obj.__dict__[attr_name]} is not type {data_type}"),
    )
    return obj.__dict__[attr_name]



from typing import Any, Dict, Iterable, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
from triad.collections.schema import Schema
from triad.utils.rename import normalize_names

from .._utils.registry import fugue_plugin
from .dataframe import AnyDataFrame, DataFrame, as_fugue_df


@fugue_plugin
def is_df(df: Any) -> bool:
    """Whether ``df`` is a DataFrame like object"""
    return isinstance(df, DataFrame)


def get_native_as_df(df: AnyDataFrame) -> AnyDataFrame:
    """Return the dataframe form of the input ``df``.
    If ``df`` is a :class:`~.DataFrame`, then call the
    :meth:`~.DataFrame.native_as_df`, otherwise, it depends on whether there is
    a correspondent function handling it.
    """
    if isinstance(df, DataFrame):
        return df.native_as_df()
    if is_df(df):
        return df
    raise NotImplementedError(f"cannot get a dataframe like object from {type(df)}")


@fugue_plugin
def get_schema(df: AnyDataFrame) -> Schema:
    """Get the schema of the ``df``

    :param df: the object that can be recognized as a dataframe by Fugue
    :return: the Schema object
    """
    return as_fugue_df(df).schema


@fugue_plugin
def as_pandas(df: AnyDataFrame) -> pd.DataFrame:
    """Convert ``df`` to a Pandas DataFrame

    :param df: the object that can be recognized as a dataframe by Fugue
    :return: the Pandas DataFrame
    """
    return as_fugue_df(df).as_pandas()


@fugue_plugin
def as_arrow(df: AnyDataFrame) -> pa.Table:
    """Convert ``df`` to a PyArrow Table

    :param df: the object that can be recognized as a dataframe by Fugue
    :return: the PyArrow Table
    """
    return as_fugue_df(df).as_arrow()


@fugue_plugin
def as_array(
    df: AnyDataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Any]:  # pragma: no cover
    """Convert df to 2-dimensional native python array

    :param df: the object that can be recognized as a dataframe by Fugue
    :param columns: columns to extract, defaults to None
    :param type_safe: whether to ensure output conforms with its schema,
        defaults to False
    :return: 2-dimensional native python array

    .. note::

        If ``type_safe`` is False, then the returned values are 'raw' values.
    """
    return as_fugue_df(df).as_array(columns=columns, type_safe=type_safe)


@fugue_plugin
def as_array_iterable(
    df: AnyDataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Any]:  # pragma: no cover
    """Convert df to iterable of native python arrays

    :param df: the object that can be recognized as a dataframe by Fugue
    :param columns: columns to extract, defaults to None
    :param type_safe: whether to ensure output conforms with its schema,
        defaults to False
    :return: iterable of native python arrays

    .. note::

        If ``type_safe`` is False, then the returned values are 'raw' values.
    """

    return as_fugue_df(df).as_array_iterable(columns=columns, type_safe=type_safe)


@fugue_plugin
def as_dict_iterable(
    df: AnyDataFrame, columns: Optional[List[str]] = None
) -> Iterable[Dict[str, Any]]:
    """Convert df to iterable of native python dicts

    :param df: the object that can be recognized as a dataframe by Fugue
    :param columns: columns to extract, defaults to None
    :return: iterable of native python dicts

    .. note::

        The default implementation enforces ``type_safe`` True
    """
    return as_fugue_df(df).as_dict_iterable(columns=columns)


@fugue_plugin
def peek_array(df: AnyDataFrame) -> List[Any]:
    """Peek the first row of the dataframe as an array

    :param df: the object that can be recognized as a dataframe by Fugue
    :return: the first row as an array
    """
    return as_fugue_df(df).peek_array()


@fugue_plugin
def peek_dict(df: AnyDataFrame) -> Dict[str, Any]:
    """Peek the first row of the dataframe as a array

    :param df: the object that can be recognized as a dataframe by Fugue
    :return: the first row as a dict
    """
    return as_fugue_df(df).peek_dict()


@fugue_plugin
def head(
    df: AnyDataFrame,
    n: int,
    columns: Optional[List[str]] = None,
    as_fugue: bool = False,
) -> AnyDataFrame:
    """Get first n rows of the dataframe as a new local bounded dataframe

    :param n: number of rows
    :param columns: selected columns, defaults to None (all columns)
    :param as_fugue: whether return a Fugue :class:`~.DataFrame`, default to
        False. If False, then if the input ``df`` is not a Fugue DataFrame
        then it will return the underlying DataFrame object.
    :return: a local bounded dataframe
    """
    res = as_fugue_df(df).head(n=n, columns=columns)
    if as_fugue or isinstance(df, DataFrame):
        return res
    return res.native_as_df()


@fugue_plugin
def alter_columns(
    df: AnyDataFrame, columns: Any, as_fugue: bool = False
) -> AnyDataFrame:
    """Change column types

    :param df: the object that can be recognized as a dataframe by Fugue
    :param columns: |SchemaLikeObject|,
        all columns should be contained by the dataframe schema
    :param as_fugue: whether return a Fugue :class:`~.DataFrame`, default to
        False. If False, then if the input ``df`` is not a Fugue DataFrame
        then it will return the underlying DataFrame object.
    :return: a new dataframe with altered columns, the order of the
        original schema will not change
    """
    return _convert_df(df, as_fugue_df(df).alter_columns(columns), as_fugue=as_fugue)


@fugue_plugin
def drop_columns(
    df: AnyDataFrame, columns: List[str], as_fugue: bool = False
) -> AnyDataFrame:
    """Drop certain columns and return a new dataframe

    :param df: the object that can be recognized as a dataframe by Fugue
    :param columns: columns to drop
    :param as_fugue: whether return a Fugue :class:`~.DataFrame`, default to
        False. If False, then if the input ``df`` is not a Fugue DataFrame
        then it will return the underlying DataFrame object.
    :return: a new dataframe removing the columns
    """
    return _convert_df(df, as_fugue_df(df).drop(columns), as_fugue=as_fugue)


@fugue_plugin
def select_columns(
    df: AnyDataFrame, columns: List[Any], as_fugue: bool = False
) -> AnyDataFrame:
    """Select certain columns and return a new dataframe

    :param df: the object that can be recognized as a dataframe by Fugue
    :param columns: columns to return
    :param as_fugue: whether return a Fugue :class:`~.DataFrame`, default to
        False. If False, then if the input ``df`` is not a Fugue DataFrame
        then it will return the underlying DataFrame object.
    :return: a new dataframe with the selected the columns
    """
    return _convert_df(df, as_fugue_df(df)[columns], as_fugue=as_fugue)


@fugue_plugin
def get_column_names(df: AnyDataFrame) -> List[Any]:  # pragma: no cover
    """A generic function to get column names of any dataframe

    :param df: the dataframe object
    :return: the column names

    .. note::

        In order to support a new type of dataframe, an implementation must
        be registered, for example

        .. code-block::python

            @get_column_names.candidate(lambda df: isinstance(df, pa.Table))
            def _get_pyarrow_dataframe_columns(df: pa.Table) -> List[Any]:
                return [f.name for f in df.schema]
    """
    return get_schema(df).names


@fugue_plugin
def rename(
    df: AnyDataFrame, columns: Dict[str, Any], as_fugue: bool = False
) -> AnyDataFrame:
    """A generic function to rename column names of any dataframe

    :param df: the dataframe object
    :param columns: the rename operations as a dict: ``old name => new name``
    :param as_fugue: whether return a Fugue :class:`~.DataFrame`, default to
        False. If False, then if the input ``df`` is not a Fugue DataFrame
        then it will return the underlying DataFrame object.
    :return: the renamed dataframe

    .. note::

        In order to support a new type of dataframe, an implementation must
        be registered, for example

        .. code-block::python

            @rename.candidate(
                lambda df, *args, **kwargs: isinstance(df, pd.DataFrame)
            )
            def _rename_pandas_dataframe(
                df: pd.DataFrame, columns: Dict[str, Any]
            ) -> pd.DataFrame:
                if len(columns) == 0:
                    return df
                return df.rename(columns=columns)
    """
    if len(columns) == 0:
        return df
    return _convert_df(df, as_fugue_df(df).rename(columns), as_fugue=as_fugue)


def normalize_column_names(df: AnyDataFrame) -> Tuple[AnyDataFrame, Dict[str, Any]]:
    """A generic function to normalize any dataframe's column names to follow
    Fugue naming rules

    .. note::

        This is a temporary solution before
        :class:`~triad:triad.collections.schema.Schema`
        can take arbitrary names

    .. admonition:: Examples

        * ``[0,1]`` => ``{"_0":0, "_1":1}``
        * ``["1a","2b"]`` => ``{"_1a":"1a", "_2b":"2b"}``
        * ``["*a","-a"]`` => ``{"_a":"*a", "_a_1":"-a"}``

    :param df: a dataframe object
    :return: the renamed dataframe and the rename operations as a dict that
        can **undo** the change

    .. seealso::

        * :func:`~.get_column_names`
        * :func:`~.rename`
        * :func:`~triad:triad.utils.rename.normalize_names`
    """
    cols = get_column_names(df)
    names = normalize_names(cols)
    if len(names) == 0:
        return df, {}
    undo = {v: k for k, v in names.items()}
    return (rename(df, names), undo)


def _convert_df(
    input_df: AnyDataFrame, output_df: DataFrame, as_fugue: bool
) -> AnyDataFrame:
    if as_fugue or isinstance(input_df, DataFrame):
        return output_df
    return output_df.native_as_df()



import pyarrow as pa
from fugue.column import col, function, lit, null, all_cols
from fugue.column.expressions import _get_column_mentions
from fugue.column.functions import coalesce
from pytest import raises
from triad import Schema, to_uuid


def test_named_col():
    assert "*" == str(all_cols())
    raises(NotImplementedError, lambda: all_cols().output_name)
    raises(NotImplementedError, lambda: all_cols().alias("x"))
    raises(NotImplementedError, lambda: all_cols().cast("long"))

    assert "a" == str(col("a"))
    assert "a" == str(col(col("a")))
    assert "ab AS xx" == str(col("ab").alias("xx"))
    assert "ab AS xx" == str(col("ab", "xx").cast(None))
    assert "CAST(ab AS long) AS xx" == str(col("ab", "xx").cast("long"))

    assert "ab AS xx" == str(col("ab").alias("xx"))

    assert "ab AS xx" == str(col("ab").alias("xx"))
    assert "CAST(ab AS long) AS xx" == str(col("ab").alias("xx").cast(int))

    raises(NotImplementedError, lambda: col([1, 2]))

    assert to_uuid(col("a")) != to_uuid(col("b"))
    assert to_uuid(col("a")) != to_uuid(col("a").alias("v"))
    assert to_uuid(col("a")) != to_uuid(col("a").cast(int))
    assert to_uuid(col("a").cast(int).alias("v")) == to_uuid(
        col("a").alias("v").cast(int)
    )

    assert "" == col("a").infer_alias().as_name
    assert "a" == str(col("a").infer_alias())
    assert "a" == col("a").cast(int).infer_alias().as_name
    c = col("a").cast(int).infer_alias()
    assert "CAST(a AS long) AS a" == str(c)
    c = col("a").cast(int).alias("x").infer_alias()
    assert "CAST(a AS long) AS x" == str(c)


def test_lit_col():
    assert "NULL" == str(lit(None))
    assert "TRUE" == str(null().is_null())
    assert "FALSE" == str(null().not_null())

    assert "'a'" == str(lit("a"))
    assert "'a\"\\'\\\\'" == str(lit("a\"'\\"))
    assert "'a' AS x" == str(lit("a", "x"))
    assert "TRUE" == str(lit("a").not_null())
    assert "FALSE" == str(lit("a").is_null())

    assert "1.1" == str(lit(1.1))
    assert "11" == str(lit(11))
    assert "TRUE" == str(lit(True))
    assert "FALSE" == str(lit(False))

    assert "1 AS xx" == str(lit(1).alias("xx"))
    assert "'ab' AS xx" == str(lit("ab").alias("xx"))

    raises(NotImplementedError, lambda: lit([1, 2]))

    assert to_uuid(lit("a")) != to_uuid(col("a"))
    assert to_uuid(lit(1)) != to_uuid(lit("1"))
    assert to_uuid(null()) == to_uuid(null())
    assert to_uuid(null()) != to_uuid(lit(1))
    assert to_uuid(lit("a")) != to_uuid(lit("a").alias("v"))
    assert to_uuid(lit("a")) != to_uuid(lit("a").cast(int))
    assert to_uuid(lit("a").cast(int).alias("v")) == to_uuid(
        lit("a").alias("v").cast(int)
    )


def test_unary_op():
    assert "-(a)" == str(-col("a"))
    assert "a" == (-col("a")).infer_alias().output_name
    assert "a" == str(+col("a"))
    assert "~(a)" == str(~col("a"))
    assert "IS_NULL(a)" == str(col("a").is_null())
    assert "NOT_NULL(a)" == str(col("a").not_null())

    assert "NOT_NULL(a) AS xx" == str(col("a").not_null().alias("xx"))
    assert "NOT_NULL(a)" == str(col("a").not_null())
    assert "NOT_NULL(a) AS xx" == str(col("a").not_null().alias("xx"))

    assert "a" == col("a").not_null().infer_alias().output_name
    assert "NOT_NULL(a) AS a" == str(col("a").not_null().infer_alias())

    assert to_uuid(col("a").not_null()) == to_uuid(col("a").not_null())
    assert to_uuid(col("a").not_null()) != to_uuid(col("a").is_null())


def test_binary_op():
    assert "+(ab,1)" == str(col("ab") + 1)
    assert "+(ab,x)" == str(col("ab") + col("x"))
    assert "+('x',a)" == str("x" + col("a"))
    assert "+('x','a')" == str("x" + lit("a"))
    assert "-(a,1)" == str(col("a") - 1)
    assert "-(1.1,a)" == str(1.1 - col("a"))
    assert "*(a,1)" == str(col("a") * 1)
    assert "*(1.1,a)" == str(1.1 * col("a"))
    assert "/(a,1)" == str(col("a") / 1)
    assert "/(1.1,a)" == str(1.1 / col("a"))

    assert "+(ab,1)" == str((col("ab") + 1))
    assert "+(ab,1) AS xx" == str((col("ab") + 1).alias("xx"))

    assert "+(ab,1) AS xx" == str((col("ab") + 1).alias("xx"))

    assert "&(a,TRUE)" == str(col("a") & True)
    assert "&(TRUE,a)" == str(True & col("a"))
    assert "&(a,FALSE)" == str(col("a") & False)
    assert "&(FALSE,a)" == str(False & col("a"))

    assert "|(a,TRUE)" == str(col("a") | True)
    assert "|(TRUE,a)" == str(True | col("a"))
    assert "|(a,FALSE)" == str(col("a") | False)
    assert "|(FALSE,a)" == str(False | col("a"))

    assert "<(a,1)" == str(col("a") < 1)
    assert "<(a,b)" == str(col("a") < col("b"))
    assert ">(a,1.1)" == str(1.1 < col("a"))
    assert "<(1.1,a)" == str(lit(1.1) < col("a"))
    assert "<=(a,1)" == str(col("a") <= 1)
    assert ">=(a,1.1)" == str(1.1 <= col("a"))
    assert ">(a,1)" == str(col("a") > 1)
    assert "<(a,1.1)" == str(1.1 > col("a"))
    assert ">=(a,1)" == str(col("a") >= 1)
    assert "<=(a,1.1)" == str(1.1 >= col("a"))

    assert "==(a,1)" == str(col("a") == 1)
    assert "==(a,1.1)" == str(1.1 == col("a"))
    assert "!=(a,1)" == str(col("a") != 1)
    assert "!=(a,1.1)" == str(1.1 != col("a"))


def test_comb():
    assert "-(+(a,*(10,b)),/(c,d))" == str(
        (col("a") + 10 * col("b")) - col("c") / col("d")
    )
    assert "|(==(a,1.1),&(&(b,~(c)),TRUE))" == str(
        (1.1 == col("a")) | col("b") & ~col("c") & True
    )


def test_function():
    expr = function("f", col("x") + col("z"), col("y"), 1, 1.1, False, "t")
    assert "f(+(x,z),y,1,1.1,FALSE,'t')" == str(expr)
    assert "f(+(x,z),y,1,1.1,FALSE,'t') AS x" == str(expr.alias("x"))


def test_coalesce():
    expr = coalesce(col("x") + col("z"), col("y"), 1, 1.1, False, "t")
    assert "COALESCE(+(x,z),y,1,1.1,FALSE,'t')" == str(expr)
    assert "COALESCE(+(x,z),y,1,1.1,FALSE,'t') AS x" == str(expr.alias("x"))


def test_get_column_mentions():
    expr = (col("a") + col("b")) * function("x", col("b"), a=col("c"), b=lit(1))
    assert set(["a", "b", "c"]) == set(_get_column_mentions(expr))


def test_schema_inference():
    schema = Schema("a:int,b:str,c:bool,d:double")
    assert pa.int32() == col("a").infer_type(schema)
    assert pa.int32() == (-col("a")).infer_type(schema)
    assert pa.int64() == (-col("a")).cast(int).infer_type(schema)
    assert pa.int64() == (-col("a").cast(int)).infer_type(schema)
    assert pa.string() == col("b").infer_type(schema)
    assert (-col("b")).infer_type(schema) is None
    assert (~col("b")).infer_type(schema) is None
    assert pa.bool_() == col("c").infer_type(schema)
    assert pa.bool_() == (~col("c")).alias("x").infer_type(schema)
    assert pa.float64() == col("d").infer_type(schema)
    assert pa.float64() == (-col("d").alias("x")).infer_type(schema)
    assert col("x").infer_type(schema) is None
    assert pa.string() == col("x").cast(str).infer_type(schema)
    assert all_cols().infer_type(schema) is None

    assert pa.bool_() == (col("a") < col("d")).infer_type(schema)
    assert pa.bool_() == (col("a") > col("d")).infer_type(schema)
    assert pa.bool_() == (col("a") <= col("d")).infer_type(schema)
    assert pa.bool_() == (col("a") >= col("d")).infer_type(schema)
    assert pa.bool_() == (col("a") == col("d")).infer_type(schema)
    assert pa.bool_() == (col("a") != col("d")).infer_type(schema)
    assert pa.bool_() == (~(col("a") != col("d"))).infer_type(schema)
    assert pa.int64() == (~(col("a") != col("d"))).cast(int).infer_type(schema)

    assert (col("a") - col("d")).infer_type(schema) is None

    assert pa.int64() == lit(1).infer_type(schema)
    assert pa.string() == lit("a").infer_type(schema)
    assert pa.bool_() == lit(False).infer_type(schema)
    assert pa.string() == lit(False).cast(str).infer_type(schema)
    assert pa.float64() == lit(2.2).infer_type(schema)
    assert null().infer_type(schema) is None
    assert pa.string() == null().cast(str).infer_type(schema)

    assert function("a", col("a").cast("int")).infer_type(schema) is None
    assert pa.string() == function("a", col("a").cast("int")).cast(str).infer_type(
        schema
    )



import pandas as pd
from fugue import FugueWorkflow
import sys
import traceback
from fugue.constants import (
    FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE,
    FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE,
)


def test_runtime_exception():
    if sys.version_info < (3, 7):
        return

    def tr(df: pd.DataFrame) -> pd.DataFrame:
        raise Exception

    def show(df):
        df.show()

    dag = FugueWorkflow()
    df = dag.df([[0]], "a:int")
    df = df.transform(tr, schema="*")
    show(df)

    try:
        dag.run()
    except Exception:
        assert len(traceback.extract_tb(sys.exc_info()[2])) < 10

    dag = FugueWorkflow({FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE: False})
    df = dag.df([[0]], "a:int")
    df = df.transform(tr, schema="*")
    show(df)

    try:
        dag.run("native")
    except Exception:
        assert len(traceback.extract_tb(sys.exc_info()[2])) > 10

    dag = FugueWorkflow({FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE: ""})
    df = dag.df([[0]], "a:int")
    df = df.transform(tr, schema="*")
    show(df)

    try:
        dag.run("native")
    except Exception:
        assert len(traceback.extract_tb(sys.exc_info()[2])) > 10


def test_modified_exception():
    if sys.version_info < (3, 7):
        return

    def tr(df: pd.DataFrame) -> pd.DataFrame:
        raise Exception

    def show(df):
        df.show()

    def tt(df):
        __modified_exception__ = NotImplementedError()
        return df.transform(tr, schema="*")

    dag = FugueWorkflow()
    df = dag.df([[0]], "a:int")
    df = tt(df)
    show(df)

    try:
        dag.run()
    except Exception as ex:
        assert isinstance(ex.__cause__, NotImplementedError)



from typing import Any

from fugue import ArrowDataFrame, DataFrame, LocalDataFrame
from fugue.plugins import as_fugue_dataset, as_local_bounded
from fugue_ibis import IbisDataFrame, IbisTable


class MockDuckDataFrame(IbisDataFrame):
    def to_sql(self) -> str:
        return str(self.native.compile())

    def _to_new_df(self, table: IbisTable, schema: Any = None) -> DataFrame:
        return MockDuckDataFrame(table, schema=schema)

    def _to_local_df(self, table: IbisTable, schema: Any = None) -> LocalDataFrame:
        return ArrowDataFrame(table.execute(), schema=schema)

    def _to_iterable_df(self, table: IbisTable, schema: Any = None) -> LocalDataFrame:
        return self._to_local_df(table, schema=schema)


# should also check the df._findbackend is duckdb
@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, IbisTable))
def _ibis_as_fugue(df: IbisTable, **kwargs: Any) -> bool:
    return MockDuckDataFrame(df, **kwargs)


# should also check the df._findbackend is duckdb
@as_local_bounded.candidate(lambda df, **kwargs: isinstance(df, IbisTable))
def _ibis_as_local(df: IbisTable, **kwargs: Any) -> bool:
    return df.execute()
